import os

import pytest

from isqtools.isqc_path import _IsqcPath, get_isqc_path, set_isqc_path


def test_initialization():
    """Test that the _IsqcPath instance initializes with an empty path."""
    isqc_path_instance = _IsqcPath()
    assert isqc_path_instance.get_path() == ""


def test_set_and_get_path():
    """Test setting and getting the path."""
    isqc_path_instance = _IsqcPath()
    test_path = "~/test_path"
    expanded_path = os.path.expanduser(test_path)

    # Set the path
    isqc_path_instance.set_path(test_path)
    assert isqc_path_instance.get_path() == expanded_path

    # Test with a different path
    another_path = "/another/path"
    isqc_path_instance.set_path(another_path)
    assert isqc_path_instance.get_path() == another_path


def test_path_expansion():
    """Test that paths are expanded correctly using os.path.expanduser."""
    test_path = "~/expanded_path"
    expanded_path = os.path.expanduser(test_path)

    # Set the path
    set_isqc_path(test_path)

    # Verify the path is expanded
    assert get_isqc_path() == expanded_path


def test_empty_path():
    """Test setting and getting an empty path."""
    set_isqc_path("")
    assert get_isqc_path() == ""


def test_non_string_path():
    """Test that setting a non-string path raises an error."""
    with pytest.raises(TypeError):
        set_isqc_path(123)  # Non-string input should raise an error



import tempfile
from pathlib import Path

import numpy as np

from isqtools import IsqCircuit
from isqtools.backend import NumpyBackend

SIMPLE_ISQ = """
import std;

qbit q[2];

procedure main() {
    H(q[0]);
    H(q[1]);
    M(q[0]);
    M(q[1]);
}
"""


def test_numpy_measure():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(SIMPLE_ISQ)
        backend = NumpyBackend()
        qc = IsqCircuit(
            file=str(temp_file_path),
            backend=backend,
            sample=False,
        )
        results = np.array([0.25, 0.25, 0.25, 0.25])
        np.testing.assert_allclose(qc.measure(), results)


def test_numpy_sample():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(SIMPLE_ISQ)
        backend = NumpyBackend()
        qc = IsqCircuit(
            file=str(temp_file_path),
            backend=backend,
            sample=True,
            shots=1024,
        )
        assert qc.measure(rng_seed=10) == {"11": 260, "00": 266, "10": 248, "01": 250}



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Autograd backend.
Autograd can automatically differentiate native Python and Numpy code.
It can handle a large subset of Python's features, including loops, ifs,
recursion and closures, and it can even take derivatives of derivatives
of derivatives. It supports reverse-mode differentiation (a.k.a.
backpropagation), which means it can efficiently take gradients of
scalar-valued functions with respect to array-valued arguments, as well as
forward-mode differentiation, and the two can be composed arbitrarily. The main
intended application of Autograd is gradient-based optimization.
For more information, check out: https://github.com/HIPS/autograd
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from .state_vector_simulator import StateVectorSimulator

try:
    import autograd.numpy as anp
except ModuleNotFoundError:
    pass

Tensor = Any


class AutogradBackend(StateVectorSimulator):
    """This method inherits the ``StateVectorSimulator``, and uses
    autograd to perform the mathematical calculation.
    """

    def as_tensor(self, tensor: Tensor) -> Tensor:
        """Convert to the corresponding tensor."""
        return anp.array(tensor)

    def H(self) -> Tensor:
        return (
            1
            / anp.sqrt(2)
            * anp.array(
                [
                    [1, 1],
                    [1, -1],
                ],
                dtype=complex,
            )
        )

    def X(self) -> Tensor:
        return anp.array(
            [
                [0, 1],
                [1, 0],
            ],
            dtype=complex,
        )

    def Y(self) -> Tensor:
        return anp.array(
            [
                [0, -1j],
                [1j, 0],
            ],
            dtype=complex,
        )

    def Z(self) -> Tensor:
        return anp.array(
            [
                [1, 0],
                [0, -1],
            ],
            dtype=complex,
        )

    def S(self) -> Tensor:
        return anp.array(
            [
                [1, 0],
                [0, 1j],
            ],
            dtype=complex,
        )

    def T(self) -> Tensor:
        return anp.array(
            [
                [anp.exp(-1j * anp.pi / 8), 0],
                [0, anp.exp(1j * anp.pi / 8)],
            ],
            dtype=complex,
        )

    def SD(self) -> Tensor:
        return anp.array(
            [
                [1, 0],
                [0, -1j],
            ],
            dtype=complex,
        )

    def TD(self) -> Tensor:
        return anp.array(
            [
                [anp.exp(1j * anp.pi / 8), 0],
                [0, anp.exp(-1j * anp.pi / 8)],
            ],
            dtype=complex,
        )

    def X2M(self) -> Tensor:
        return (
            anp.sqrt(2)
            / 2
            * anp.array(
                [
                    [1, 1j],
                    [1j, 1],
                ],
                dtype=complex,
            )
        )

    def X2P(self) -> Tensor:
        return (
            anp.sqrt(2)
            / 2
            * anp.array(
                [
                    [1, -1j],
                    [-1j, 1],
                ],
                dtype=complex,
            )
        )

    def Y2M(self) -> Tensor:
        return (
            anp.sqrt(2)
            / 2
            * anp.array(
                [
                    [1, 1],
                    [-1, 1],
                ],
                dtype=complex,
            )
        )

    def Y2P(self) -> Tensor:
        return (
            anp.sqrt(2)
            / 2
            * anp.array(
                [
                    [1, -1],
                    [1, 1],
                ],
                dtype=complex,
            )
        )

    def RX(self, theta: Tensor) -> Tensor:
        theta /= 2
        return anp.array(
            [
                [anp.cos(theta), -1j * anp.sin(theta)],
                [-1j * anp.sin(theta), anp.cos(theta)],
            ],
            dtype=complex,
        )

    def RY(self, theta: Tensor) -> Tensor:
        theta /= 2
        return anp.array(
            [
                [anp.cos(theta), -1 * anp.sin(theta)],
                [anp.sin(theta), anp.cos(theta)],
            ],
            dtype=complex,
        )

    def RZ(self, theta: Tensor) -> Tensor:
        theta /= 2
        return anp.array(
            [
                [anp.exp(-1j * theta), 0],
                [0, anp.exp(1j * theta)],
            ],
            dtype=complex,
        )

    def RXY(self, theta: Tensor, phi: Tensor) -> Tensor:
        theta /= 2
        return anp.array(
            [
                [anp.cos(theta), -1j * anp.exp(-1j * phi) * anp.sin(theta)],
                [-1j * anp.exp(1j * phi) * anp.sin(theta), anp.cos(theta)],
            ],
            dtype=complex,
        )

    def get_zero_state(self, qnum: int) -> Tensor:
        state = anp.zeros(1 << qnum, dtype=complex)
        state[0] = 1
        return state

    def reshape(self, state: Tensor, shape: Sequence[int]) -> Tensor:
        return anp.reshape(state, shape)

    def ravel(self, state: Tensor) -> Tensor:
        return anp.ravel(state)

    def stack(self, states: Sequence[Tensor], axis: int = 0) -> Tensor:
        return anp.stack(states, axis=axis)

    def real(self, state: Tensor) -> Tensor:
        return anp.real(state)

    def sum(self, state: Tensor, axis: int | None = None) -> Tensor:
        return anp.sum(state, axis=axis)

    def conj(self, state: Tensor) -> Tensor:
        return anp.conj(state)

    def sqrt(self, state: Tensor) -> Tensor:
        return anp.sqrt(state)

    def copy(self, state: Tensor) -> Tensor:
        return state.copy()

    def dot(self, state1: Tensor, state2: Tensor) -> Tensor:
        return anp.dot(state1, state2)



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""A connector to directly use circuits as PyTorch modules.
This part code is inspired by tensorcircuit.
"""

from __future__ import annotations

from typing import Any, Callable

try:
    import torch
    from torch import Tensor
    from torch.nn import Module, Parameter
except Exception:
    Module = object
    Tensor = Any


class TorchLayerError(Exception):
    """Torch layer exception."""


class TorchLayer(Module):
    """Apply a transformation of quantum circuits to the incoming data"""

    __constants__ = ["num_weights"]
    num_weights: int

    def __init__(
        self,
        circuit: Callable[..., Any],
        num_weights: int,
        *,
        is_vmap: bool = True,
        in_dims: tuple[int | None, int | None] = (0, None),
        initial_weights: Tensor | None = None,
    ) -> None:
        super().__init__()

        # vamp is used, pytorch > 2.0
        self.is_vmap = is_vmap
        if is_vmap:
            circuit = torch.vmap(circuit, in_dims=in_dims)
            self.in_dims = in_dims

        # initialize weights
        if isinstance(initial_weights, Tensor):
            self.weights = Parameter(initial_weights)
        else:
            self.weights = Parameter(torch.randn(num_weights))

        self.circuit = circuit
        self.num_weights = num_weights

    def forward(self, inputs: Tensor) -> Tensor:
        return self.circuit(inputs, self.weights)

    def extra_repr(self) -> str:
        """Nice print this layer."""
        if self.is_vmap:
            return (
                f"num_weights={self.num_weights}, "
                f"is_vmap={self.is_vmap}, "
                f"vmap_in_dims={self.in_dims}"
            )
        else:
            return f"num_weights={self.num_weights}, is_vmap={self.is_vmap}"



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Quantum circuits are at the heart of how quantum computers operate. This
package implements tools related to quantum circuits.
"""

from .isq_circuit import IsqCircuit



import numpy as np
from openfermion.transforms import get_fermion_operator
from pyscf import gto, scf

from isqtools.chem.utils import (
    get_molecular_hamiltonian,
    openfermion_to_pauligates,
    parity,
)

bond_length = 1.30

atom = f"""
 H 0 0 0
 Li 0 0 {bond_length}
"""

basis = "sto-3g"
charge = 0
spin = 0
mymol = gto.M(atom=atom, basis=basis, charge=charge, spin=spin, verbose=0)
nao = int(mymol.nao)
myhf = scf.RHF(mymol).run()
molecular_hamiltonian = get_molecular_hamiltonian(myhf)
fermion_hamiltonian = get_fermion_operator(molecular_hamiltonian)
qubit_hamiltonian = parity(fermion_hamiltonian, nao * 2, nao)
qubit_hamiltonian.compress()
# print("Qubit hamiltonian:", qubit_hamiltonian)
coeffs, gates_group = openfermion_to_pauligates(qubit_hamiltonian)
print("Coefficients:", coeffs)
print("Gates:", gates_group)



import pytest

pyscf = pytest.importorskip("pyscf")


from pyscf import gto


def test_charge():
    bond_length = 0.75

    atom = f"""
     H 0 0 0
     H 0 0 {bond_length}
    """

    basis = "sto-3g"
    charge = 0
    spin = 0
    mymol = gto.M(atom=atom, basis=basis, charge=charge, spin=spin, verbose=3)
    myhf = mymol.HF()
    myhf.kernel()
    myhf.analyze()



import pytest

torch = pytest.importorskip("torch")


import random
import tempfile
from pathlib import Path

import numpy as np
import torch
import torch.optim as optim

from isqtools import IsqCircuit
from isqtools.backend import TorchBackend
from isqtools.neural_networks import TorchLayer


@pytest.fixture(scope="module", autouse=True)
def set_seed():
    seed = 222
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


@pytest.fixture(scope="module")
def qc():
    with tempfile.TemporaryDirectory() as temp_dir:
        script_dir = Path(__file__).resolve().parent
        file_path = script_dir / "nn_basic.isq"
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        temp_file_path.write_text(file_path.read_text())

        if not temp_file_path.exists():
            raise FileNotFoundError(f"File not found: {temp_file_path}")

        backend = TorchBackend()
        isq_circuit = IsqCircuit(
            file=str(temp_file_path),
            backend=backend,
            sample=False,
        )
        yield isq_circuit


@pytest.fixture(scope="module")
def circuit(qc):
    def _circuit(inputs, weights):
        param = {
            "inputs": inputs,
            "weights": weights,
        }
        return qc.pauli_measure(**param)

    return _circuit


@pytest.fixture(scope="module")
def qnn(circuit):
    weights = torch.randn(24)
    return TorchLayer(
        circuit=circuit,
        is_vmap=False,
        num_weights=24,
        initial_weights=weights,
    )


def test_nn_run(circuit):
    inputs = torch.randn(4)
    weights = torch.randn(24)
    expected_result = np.array([-0.127469], dtype=np.float32)

    result = circuit(inputs, weights).detach().cpu().numpy()
    np.testing.assert_allclose(result, expected_result[0], atol=1e-6)


def test_qnn(qnn):
    inputs = torch.randn(4)
    expected_result = np.array([0.03218], dtype=np.float32)

    result = qnn(inputs).detach().cpu().numpy()
    np.testing.assert_allclose(result, expected_result[0], atol=1e-6)


def test_opt(qnn):
    inputs = torch.randn(4)
    optimizer = optim.SGD(qnn.parameters(), lr=0.05)

    for i in range(50):
        measurement = qnn(inputs)
        measurement.backward()
        optimizer.step()
        optimizer.zero_grad()
    assert measurement.item() < 0.9



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Isq quantum circuit.
Use ``isqc`` to generate qcis files, and generate quantum circuits by reading
qcis files. IsqCircuit is our recommended way to build quantum circuits.
"""

from __future__ import annotations

import os
from collections.abc import Sequence
from typing import Any

from isqtools.backend import AbstractBackend, NumpyBackend
from isqtools.utils import gen_qcis_from_so, isqc_compile

from .abstract_circuit import AbstractCircuit

Tensor = Any


class IsqCircuitError(Exception):
    """IsQ quantum circuits error."""


class IsqCircuit(AbstractCircuit):
    """IsQ quantum circuit class.

    Attributes:
        backend: The backend on which the task runs, such as a simulator or hardware.
        sample: Whether to enable sampling mode. For simulators, supports
            ``sample`` or ``probs`` modes. For hardware, only sampling mode
            is supported.
        shots: Specifies the number of shots when sampling mode is enabled.
        qcis: Qcis instruction set for quantum circuits.

    """

    def __init__(
        self,
        file: str | None = None,
        backend: AbstractBackend | None = None,
        sample: bool = False,
        shots: int = 100,
        int_param: Sequence[int] | int | None = None,
        double_param: Sequence[float] | float | None = None,
        **kw,
    ) -> None:
        """Initialization of isq quantum circuits.

        Args:
            file: The file path. It is recommended to use full paths, be
                  careful when using relative paths.
                  1) When the file extension is ``.isq``, the isq file will be
                     compiled into qcis.
                  2) When the extension of the file is ``qcis``, read the file
                     directly.
                  3) When the extension of the file is ``so``, it will be
                     called ``gen_qcis_from_so`` to process the ``so`` file
                     to generate qcis.
            backend: The backend on which the task runs can be a simulator or
                     hardware.
            sample: Whether to enable sampling mode. When the backend is a
                    simulator, you can ``sample`` or ``probs``. If the backend
                    is hardware, you can only enable the sampling mode.
            shots: When the sampling mode is enabled, this parameter can
                   specify the shots number.
            int_param: Int data when compiling qcis.
            double_param: Double data when compiling qcis.
            **kw: Optional parameters.

        Raises:
            IsqCircuitError: Check the extension of the ``file``.

        """

        if file is not None:
            if file.endswith(".isq"):
                # compile every time you build the class
                isqc_compile(
                    file=file,
                    target="qcis",
                    int_param=int_param,
                    double_param=double_param,
                )
                qcis_path = os.path.splitext(file)[0] + ".qcis"

            elif file.endswith(".qcis"):
                qcis_path = file

            elif file.endswith(".so"):
                # compile once
                # this is an advanced, not recommended
                gen_qcis_from_so(
                    file=file,
                    int_param=int_param,
                    double_param=double_param,
                )
                qcis_path = os.path.splitext(file)[0] + ".qcis"

            else:
                raise IsqCircuitError(
                    f"`{file}`'s format is not supported"
                    "Please use `.isq` or `.qcis` format of circuits."
                    "Both absolute `/home/username/file.isq` and"
                    "relative paths `file.isq` are supported,"
                    "but this kind of path `./file.isq` is not supported"
                )

            with open(qcis_path, "r") as qcis:
                lines = (line.rstrip() for line in qcis)
                lines = (line for line in lines if line)
                self.qcis = "\n".join(lines)
        else:
            self.qcis = kw["qcis"]

        if backend is None:
            backend = NumpyBackend()

        super().__init__(
            backend=backend,
            sample=sample,
            shots=shots,
        )

    def measure(self, **kw) -> dict[str, int] | Tensor:
        """Taking measurements on quantum circuits.

        Args:
            **kw: Corresponding to the parameters in qcis and the ``param`` in
                  isq file.

        Returns:
            Depending on whether to sample or not, the measurement result
            is obtained.

        """
        if not self.sample:
            measure_result = self.backend.probs(self.qcis, **kw)
        else:
            measure_result = self.backend.sample(self.qcis, self.shots, **kw)
        return measure_result

    def pauli_measure(self, **kw) -> Tensor:
        """Taking Pauli measurements on quantum circuits.

        Args:
            **kw: Corresponding to the parameters in qcis and the ``param`` in
                  isq file.

        Returns:
            Depending on whether to sample or not, the measurement result
            is obtained. The measurement result will get the final result
            according to the rules of Pauli operation. This measurement method
            needs to implement the corresponding Pauli measurement in the isq
            file.

        """
        if self.sample:
            measure_result_dict = self.measure(**kw)
            result = 0
            for res_index, frequency in measure_result_dict.items():
                parity = (-1) ** (res_index.count("1") % 2)
                # e.g. {"011": 222}, to count "1"
                result += parity * frequency / self.shots
            return result
        else:
            measure_result_array = self.measure(**kw)
            parity = [
                (-1) ** (str(bin(int(index))).count("1") % 2)
                for index in range(len(measure_result_array))
            ]
            return self.backend.dot(
                measure_result_array,
                self.backend.as_tensor(parity),
            )

    def state(self, **kw) -> Tensor:
        """Get the simulated quantum state.

        Args:
            **kw: Corresponding to the parameters in qcis and the ``param`` in
                  isq file.

        Returns:
            State vector.

        Raises:
            IsqCircuitError: This method can only be used in non-sample.

        """
        if not self.sample:
            return self.backend.state(self.qcis, **kw)
        else:
            raise IsqCircuitError(
                "The quantum state cannot be obtained during the sample."
            )



import pytest

pyscf = pytest.importorskip("pyscf")
openfermion = pytest.importorskip("openfermion")


import numpy as np
from openfermion.linalg import get_ground_state, get_sparse_operator
from openfermion.transforms import get_fermion_operator
from pyscf import gto, scf

from isqtools.chem.utils import get_molecular_hamiltonian, parity


def test_pes():
    energies = []

    bond_lengths = [0.45, 0.55, 0.65, 0.75, 0.85, 0.95, 1.05]
    for bond_length in bond_lengths:

        atom = f"""
         H 0 0 0
         H 0 0 {bond_length}
        """

        basis = "sto-3g"
        charge = 0
        spin = 0
        mymol = gto.M(atom=atom, basis=basis, charge=charge, spin=spin, verbose=0)
        nao = int(mymol.nao)
        myhf = scf.RHF(mymol).run()
        molecular_hamiltonian = get_molecular_hamiltonian(myhf)
        fermion_hamiltonian = get_fermion_operator(molecular_hamiltonian)
        qubit_hamiltonian = parity(fermion_hamiltonian, nao * 2, nao)
        qubit_hamiltonian.compress()
        sparse_operator = get_sparse_operator(qubit_hamiltonian)
        energy, _ = get_ground_state(sparse_operator)
        energies.append(energy)
    # print("PES:", energies)
    np.testing.assert_allclose(
        np.array(energies),
        np.array(
            [
                -0.9984155960160171,
                -1.0926299067451068,
                -1.1299047843229142,
                -1.1371170673457303,
                -1.1283618784581126,
                -1.1113394177361482,
                -1.090342176512763,
            ]
        ),
    )



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""A Quantum Neural Network implemented by autograd.
"""

from __future__ import annotations

import logging
from typing import Any, Callable

import numpy as np

from .neural_network import NeuralNetwork

try:
    import autograd.numpy as anp
    from autograd import jacobian
except ImportError as e:
    logging.error(f"Failed to import autograd modules: {e}")


Tensor = Any


class QNNAutograd(NeuralNetwork):
    """Use autograd's automatic differentiation (autodiff) to implement the
    ``forward`` and ``backward``of the neural network.
    """

    def __init__(
        self,
        circuit: Callable[[Tensor, Tensor], Tensor],
        num_inputs: int,
        num_weights: int,
        output_shape: int | tuple[int, ...],
        input_gradients: bool = True,
        sparse: bool = False,
    ) -> None:
        super().__init__(
            num_inputs=num_inputs,
            num_weights=num_weights,
            sparse=sparse,
            output_shape=output_shape,
            input_gradients=input_gradients,
        )

        self.circuit = circuit

    def _forward(
        self,
        inputs: Tensor,
        weights: Tensor,
    ) -> Tensor:
        """The implementation of forward calculation."""
        inputs = anp.array(inputs)
        weights = anp.array(weights)

        output_data = [self.circuit(input, weights) for input in inputs]
        return np.array(output_data).reshape(-1, len(inputs))

    def _backward(
        self,
        inputs: Tensor,
        weights: Tensor,
    ) -> Tensor:
        """The implementation of backward calculation."""
        inputs = anp.array(inputs)
        weights = anp.array(weights)

        inputs_grad = np.array(
            [jacobian(self.circuit, 0)(input, weights) for input in inputs]
        )
        inputs_shape = inputs_grad.shape

        weights_grad = np.array(
            [jacobian(self.circuit, 1)(input, weights) for input in inputs]
        )
        weights_shape = weights_grad.shape

        if not self._input_gradients:
            return None, np.array(weights_grad)

        return (
            inputs_grad.reshape(-1, *inputs_shape),
            weights_grad.reshape(-1, *weights_shape),
        )



import os


class _IsqcPath:
    """Singleton class to manage the default ``isqc`` path."""

    __slots__ = ("default_isqc_path",)

    def __init__(self):
        self.default_isqc_path = ""

    def get_path(self) -> str:
        """Get the current default ``isqc`` path."""
        return self.default_isqc_path

    def set_path(self, isqc_path: str) -> None:
        """Set the default ``isqc`` path.

        Args:
            isqc_path (str): The path to set as the default ``isqc`` path.
        """
        self.default_isqc_path = os.path.expanduser(isqc_path)


_default_isqc_path = _IsqcPath()


def get_isqc_path() -> str:
    """Get the current default ``isqc`` path.

    Returns:
        str: The current default ``isqc`` path.
    """
    return _default_isqc_path.get_path()


def set_isqc_path(isqc_path: str) -> None:
    """Set the default ``isqc`` path.

    Args:
        isqc_path (str): The path to set as the default ``isqc`` path.
    """
    _default_isqc_path.set_path(isqc_path)



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Abstract quantum circuit.
The core of current quantum circuits is the Qcis instruction set. The qcis file
compiled by the isq compiler (icqc) is the main quantum circuit construction
method, and other methods are not recommended.
"""

from abc import ABC, abstractmethod

import numpy as np

from isqtools.backend import AbstractBackend


class AbstractCircuit(ABC):
    """Abstract quantum circuit.

    Attributes:
        backend: The backend on which the task runs can be a simulator or
                 hardware.
        sample: Whether to enable sampling mode. When the backend is a
                simulator, you can ``sample`` or ``probs``. If the backend is
                hardware, you can only enable the sampling mode.
        shots: When the sampling mode is enabled, this parameter can specify
               the shots number.

    """

    def __init__(
        self,
        backend: AbstractBackend,
        sample: bool,
        shots: int = 100,
    ) -> None:
        self.backend = backend
        self.sample = sample
        self.shots = shots

    @abstractmethod
    def measure(self, **kw):
        """Taking measurements on quantum circuits."""

    @abstractmethod
    def pauli_measure(self, **kw):
        """Taking Pauli measurements on quantum circuits."""

    def __repr__(self) -> str:
        """Returns the Qcis instruction set for a quantum circuit."""
        return self.qcis

    __str__ = __repr__

    def dict2array(
        self,
        result_dicts: dict,
    ) -> np.ndarray:
        """Transfer measurement results from Dict to Array.

        Args:
            result_dicts: Use a dict to represent measurements, for example:
                          ``{"00": 10, "01": 25, "10": 30, "11": 35}``.

        Returns:
            Use array to represent the measurement results, arranged in binary
            order, for example: ``[0.1, 0.25, 0.3, 0.35]``.

        """
        shots = self.shots
        len_array = len(list(result_dicts.keys())[0])
        results_arrays = np.zeros(2**len_array)

        for bin_idx, freq in result_dicts.items():
            results_arrays[int(bin_idx, 2)] = float(freq / shots)
        return results_arrays

    def sort_dict(
        self,
        result_dicts: dict,
    ) -> dict:
        """After the python >= 3.6, the dict will save the order. Use this
        method to sort the measurement results dictionary in binary.
        """
        return dict(
            sorted(
                result_dicts.items(),
                key=lambda item: int(item[0], 2),
            )
        )



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Quantum hardware operations are performed by submitting the
QCIS instruction set. This file is the basis for hardware.
"""

from __future__ import annotations

import logging
from collections import defaultdict
from typing import NoReturn

from .abstract_backend import AbstractBackend


class HardwareBackendError(Exception):
    """Error for hardware running."""


class HardwareBackend(AbstractBackend):
    """Hardware running base."""

    def probs(self) -> NoReturn:
        """Running a circuit on hardware is different from a simulator,
        and it is impossible to obtain an accurate probability distribution.

        Raises:
            HardwareBackendError: ``probs`` cannot run on the hardware,
            please use the ``sample``.

        """
        raise HardwareBackendError(
            "To run on hardware, "
            "you must use the `sample` method and specify `shots`. "
            "`probs` is not supported."
        )

    def sample(
        self,
        data: str,
        shots: int = 100,
        **kwargs,
    ) -> dict[str, int]:
        """Sampling quantum circuits using hardware. This is an abstraction
        over the concrete method.

        Args:
            data: QCIS strings, This represents the information of the
                  quantum circuit.
            shots: Shots numbers.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            This method returns a dictionary, the key in the dictionary is
            the quantum state, and the value is the shots of the quantum state.

        """
        raise NotImplementedError


class TaskType:
    """Type of hardware."""

    QCIS = "QCIS"
    QCIS2 = "QCIS2"
    ARCLIGHT = "ARCLIGHT"
    # AWS = "AWS"
    # SCQ = "SCQ"


class TaskState:
    """Types of hardware tasks."""

    WAIT = "WAITING"
    COMPLETE = "COMPLETED"
    FAIL = "FAILED"
    CANCEL = "CANCELLED"


class ISQTask:
    """isq hardware task interface, used to execute hardware tasks."""

    NO_RESULT_TERMINAL_STATES = {TaskState.FAIL, TaskState.CANCEL}

    def __init__(
        self,
        task_id: int | str,
        task_type: str,
        state: str,
        device: HardwareBackend,
        shots: int,
        **kwargs,
    ) -> None:
        self._id = task_id
        self._type = task_type
        self._state = state
        self._res = {}
        self._device = device
        self.logger = logging.getLogger(__name__)
        if "logger" in kwargs:
            self.logger = kwargs["logger"]
        self.shots = shots

    @property
    def state(self) -> str:
        """Get the status of the task.

        Returns:
            Returns four different states:
            1) WAITING.
            2) COMPLETED.
            3) FAILED.
            4) CANCELLED.

        """
        return self._state

    def result(self) -> dict:
        """Query the result of the task according to the id
        of the task.

        Returns:
            This method returns a dictionary, which is the same fotmat
            as ``sample``.

        """
        if self._state in self.NO_RESULT_TERMINAL_STATES:
            return {}

        elif self._state == TaskState.COMPLETE:
            return self._res

        if self._type == TaskType.QCIS:
            # When using qcis hardware, the id of the task is used to
            # get the result of the task.
            measure_result = self._device._account.query_experiment(
                self._id,
                max_wait_time=self._device._max_wait_time,
            )
            if measure_result:
                # measure_result example (shots = 10):
                # [{'resultStatus': [[14, 9, 15], [1, 1, 1], [1, 0, 0],
                # [1, 1, 0], [1, 1, 1], [1, 0, 0], [0, 1, 1], [1, 1, 0],
                # [0, 0, 1], [0, 0, 1], [1, 0, 0]], 'probability': {'000':
                # 0.0, '011': 0.2062521227977389, '110': 0.09061964381995466,
                # '001': 0.2800637719904287, '100': 0.18968664161204915, '111':
                # 0.23337781977982858, '101': 0.0, '010': 0.0},
                # 'experimentTaskId': '1717068093878177792'}]
                self.logger.info("Task executed successfully!")
                self._state = TaskState.COMPLETE
                if self._device.is_prob == "0":
                    self._res = {
                        k: v * self.shots
                        for k, v in measure_result[0]["probability"].items()
                    }
                elif self._device.is_prob == "1":
                    answer = defaultdict(int)
                    # defaultdict is used for data processing.
                    for t in measure_result[0]["resultStatus"][1:]:
                        x = "".join([str(v) for v in t])
                        answer[x] += 1
                    self._res = dict(answer)
                elif self._device.is_prob == "2":
                    self._res = measure_result[0]["resultStatus"][1:]
            else:
                self.logger.info("Task execution, please wait.")

        if self._type == TaskType.QCIS2:
            # When using qcis hardware, the id of the task is used to
            # get the result of the task.
            measure_result = self._device._account.query_experiment(
                self._id,
                max_wait_time=self._device._max_wait_time,
            )
            if measure_result:
                # measure_result example (shots = 10):
                # [{'probability': {'0': 0.7002964101438138,
                # '1': 0.2997035898561862}, # 'experimentTaskId':
                # '3822987317073968ustc00000829969',
                # 'results': [[27], [0], [0], [1], [0], [1], [1], [0], [0],
                # [0], [0]]}]
                self.logger.info("Task executed successfully!")
                self._state = TaskState.COMPLETE
                if self._device.is_prob:
                    self._res = {
                        k: v * self.shots
                        for k, v in measure_result[0]["probability"].items()
                    }
                else:
                    answer = defaultdict(int)
                    # defaultdict is used for data processing.
                    for t in measure_result[0]["results"][1:]:
                        x = "".join([str(v) for v in t])
                        answer[x] += 1
                    self._res = dict(answer)
            else:
                self.logger.info("Task execution, please wait.")

        if self._type == TaskType.ARCLIGHT:
            measure_result = self._device._get_result(self._id)
            if measure_result:
                self._state = TaskState.COMPLETE
                self._res = measure_result.get("data")

        return self._res



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Quantum neural network package.
"""

from .neural_network import NeuralNetwork
from .qnn_autograd import QNNAutograd
from .qnn_param_shift import QNNParamShift
from .torch_layer import TorchLayer
from .torch_wrapper import TorchWrapper



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Some useful functions of quantum chemistry.
This file is inspired by TenCirChem and OpenFermion.
TenCirChem: https://github.com/tencent-quantum-lab/TenCirChem
OpenFermion: https://quantumai.google/openfermion
"""

from __future__ import annotations

from itertools import product
from typing import Any

import numpy as np

try:
    from pyscf import ao2mo
    from pyscf.gto import M
    from pyscf.mcscf import CASCI
    from pyscf.scf.hf import RHF
except Exception:
    RHF = Any
    M = Any

try:
    from openfermion import (
        FermionOperator,
        QubitOperator,
        binary_code_transform,
        parity_code,
    )
    from openfermion.ops.representations import InteractionOperator
    from openfermion.utils import hermitian_conjugated
except Exception:
    InteractionOperator = Any
    FermionOperator = Any
    QubitOperator = Any

TOLERANCE = 1e-8


def spin_orb_from_int(
    int1e: np.ndarray,
    int2e: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """Spin orbital coefficients obtained from electron integrals.

    Args:
        int1e: Single electron integration.
        int2e: Double electron integration.

    Returns:
        One body coefficients and two body coefficients.

    """

    n_orb = int1e.shape[0]  # orbitals
    if int1e.shape != (n_orb, n_orb):
        raise ValueError(f"Invalid one-body integral array shape: {int1e.shape}")
    if int2e.shape != (n_orb, n_orb, n_orb, n_orb):
        raise ValueError(f"Invalid two-body integral array shape: {int2e.shape}")

    n_sorb = n_orb * 2  # spin orbitals
    one_body_coefficients = np.zeros(
        (
            n_sorb,
            n_sorb,
        )
    )
    two_body_coefficients = np.zeros(
        (
            n_sorb,
            n_sorb,
            n_sorb,
            n_sorb,
        )
    )

    one_body_coefficients[
        :n_orb,
        :n_orb,
    ] = one_body_coefficients[n_orb:, n_orb:] = int1e

    for p, q, r, s in product(range(n_sorb), repeat=4):
        # a_p^\dagger a_q^\dagger a_r a_s
        if ((p < n_orb) == (s < n_orb)) and ((q < n_orb) == (r < n_orb)):
            # note the different orders of the indices
            two_body_coefficients[p, q, r, s] = int2e[
                p % n_orb,
                s % n_orb,
                q % n_orb,
                r % n_orb,
            ]

    # Truncate
    one_body_coefficients[np.absolute(one_body_coefficients) < TOLERANCE] = 0.0
    two_body_coefficients[np.absolute(two_body_coefficients) < TOLERANCE] = 0.0

    return one_body_coefficients, two_body_coefficients


def canonical_mo_coeff(mo_coeff: np.ndarray) -> np.ndarray:
    """Make the first large element positive all elements smaller than 1e-5 is
    highly unlikely (at least 1e10 basis).

    Args:
        mo_coeff: Molecular orbital coefficient.

    Returns:
        Molecular orbital coefficient after truncation.

    """
    largest_elem_idx = np.argmax(1e-5 < np.abs(mo_coeff), axis=0)
    largest_elem = mo_coeff[
        (
            largest_elem_idx,
            np.arange(len(largest_elem_idx)),
        )
    ]

    return mo_coeff * np.sign(largest_elem).reshape(1, -1)


def get_int_from_hf(
    hf: RHF,
    active_space: tuple = None,
) -> tuple[np.ndarray, np.ndarray, float]:
    """Use pyscf to get one- and two-electron integrals, as well as electron
    energies from Hartree-Fock.

    Args:
        hf: Hartree-Fock from ``pyscf``.
        active_space: active space for CASCI calculation.

    Returns:
        Electronic integrals and electronic energy.

    """

    if not isinstance(hf, RHF):
        raise TypeError(f"hf object must be RHF class, got {type(hf)}")

    m = hf.mol
    assert hf.mo_coeff is not None
    hf.mo_coeff = canonical_mo_coeff(hf.mo_coeff)
    if active_space is None:
        nelecas = m.nelectron
        ncas = m.nao
    else:
        nelecas, ncas = active_space
    casci = CASCI(hf, ncas, nelecas)
    int1e, e_core = casci.get_h1eff()
    int2e = ao2mo.restore("s1", casci.get_h2eff(), ncas)

    return int1e, int2e, e_core


def get_int_from_mol(
    mol: M,
    C: np.array,
) -> tuple[np.ndarray, np.ndarray, float]:
    """Use pyscf to get one- and two-electron integrals, as well as electron
    energies from ``mol``. Do not use Hartree-Fock molecular coefficients as a
    starting point.

    Args:
        mol: molecule(mol) from ``pyscf``.
        C: Custom orbital coefficients.

    Returns:
        Electronic integrals and electronic energy.

    """

    e_core = mol.energy_nuc()
    AO = mol.intor("int2e", aosym=1)
    h = mol.get_hcore()
    # h = mol.intor("int1e_kin") + mol.intor("int1e_nuc")
    # C = hf.mo_coeff
    int1e = C.T @ h @ C
    # this can be optimized by ``opt_einsum``
    int2e = np.einsum("uvkl, up, vq, kr, ls -> pqrs", AO, C, C, C, C)
    return int1e, int2e, e_core


def get_molecular_hamiltonian(
    hf: RHF,
) -> InteractionOperator:
    """Get get molecular hamiltonian.

    Args:
        hf: Hartree-Fock from ``pyscf``.

    Returns:
        Molecular hamiltonian by ``openfermion``.

    """
    int1e, int2e, e_core = get_int_from_hf(hf)
    one_body_coefficients, two_body_coefficients = spin_orb_from_int(
        int1e,
        int2e,
    )
    return InteractionOperator(
        e_core, one_body_coefficients, 1 / 2 * two_body_coefficients
    )


def ex_op_to_fop(
    ex_op: list,
    with_conjugation: bool = False,
) -> FermionOperator:
    """Excited operators to Fermion operators.

    Args:
        ex_op: Excitation operators.
        with_conjugation: Conjugate or not.

    Returns:
        Fermion operators.

    """
    if len(ex_op) == 2:
        fop = FermionOperator(f"{ex_op[0]}^ {ex_op[1]}")
    else:
        assert len(ex_op) == 4
        fop = FermionOperator(f"{ex_op[0]}^ {ex_op[1]}^ {ex_op[2]} {ex_op[3]}")
    if with_conjugation:
        fop = fop - hermitian_conjugated(fop)
    return fop


def reverse_qop_idx(
    op: QubitOperator,
    n_qubits: int,
) -> QubitOperator:
    """Reverse qubit operators."""
    ret = QubitOperator()
    for pauli_string, value in op.terms.items():
        # internally QubitOperator assumes ascending index
        pauli_string = tuple(
            reversed(
                [
                    (
                        n_qubits - 1 - idx,
                        symbol,
                    )
                    for idx, symbol in pauli_string
                ]
            )
        )
        ret.terms[pauli_string] = value
    return ret


def reverse_fop_idx(
    op: FermionOperator,
    n_qubits: int,
) -> FermionOperator:
    """Reverse fermion operators."""
    ret = FermionOperator()
    for word, v in op.terms.items():
        word = tuple([(n_qubits - 1 - idx, symbol) for idx, symbol in word])
        ret.terms[word] = v
    return ret


def parity(
    fermion_operator: FermionOperator,
    n_modes: int,
    n_elec: int,
) -> QubitOperator:
    """Performs parity transformation.

    Args:
        fermion_operator: The fermion operator.
        n_modes: The number of spin-orbitals.
        n_elec: The number of electrons.

    Returns:
         Qubit operators.

    """
    qubit_operator = _parity(
        reverse_fop_idx(
            fermion_operator,
            n_modes,
        ),
        n_modes,
    )
    res = 0
    assert n_modes % 2 == 0
    reduction_indices = [n_modes // 2 - 1, n_modes - 1]
    phase_alpha = (-1) ** (n_elec // 2)
    for qop in qubit_operator:
        # qop example: 0.5 [Z1 X2 X3]
        pauli_string, coeff = next(iter(qop.terms.items()))
        # pauli_string example: ((1, 'Z'), (2, 'X'), (3, 'X'))
        # coeff example: 0.5
        new_pauli_string = []
        for idx, symbol in pauli_string:
            is_alpha = idx <= reduction_indices[0]
            if idx in reduction_indices:
                if symbol in ["X", "Y"]:
                    # discard this term because the bit will never change
                    continue
                else:
                    assert symbol == "Z"
                    if is_alpha:
                        coeff *= phase_alpha
                    continue
            if not is_alpha:
                idx -= 1
            new_pauli_string.append((idx, symbol))
        qop.terms = {tuple(new_pauli_string): coeff}
        res += qop
    return res


def _parity(fermion_operator, n_modes):
    return binary_code_transform(fermion_operator, parity_code(n_modes))


def openfermion_to_pauligates(qubit_hamiltonian):
    """Extract coefficients and gates of qubit hamiltonian.

    Args:
        qubit_hamiltonian: Qubit hamiltonian.

    Returns:
        Coefficients and corresponding Pauli operators.

    """
    coeffs = []
    gates_group = []
    for idx, (gate, coeff) in enumerate(qubit_hamiltonian.terms.items()):
        coeffs.append(coeff)
        if idx == 0:
            continue
        gates_group.append(gate)
    return coeffs, gates_group


def get_isq_pauli_gates(
    gates_group: list,
    num_qubits: int,
) -> list:
    """Convert the Pauli operator into a list to facilitate isq recognition.

    Args:
        gates_group: Pauli operators from ``openfermion_to_pauligates``.
        num_qubits: Number of qubits.

    Returns:
        Isq representation of Pauli operator.

    """
    pauli_gates = []
    for gates in gates_group:
        gates_temp = [0] * num_qubits
        for gate in gates:
            if gate[1] == "X":
                gates_temp[gate[0]] = 1
            elif gate[1] == "Y":
                gates_temp[gate[0]] = 2
            elif gate[1] == "Z":
                gates_temp[gate[0]] = 3
        pauli_gates.extend(gates_temp)
    return pauli_gates



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Quantum computers have a wide range of applications in chemistry. This
library mainly packages some tools used by quantum computers when performing
chemical simulations.
"""



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Simple visualization of quantum circuits. The code draws on pennylane
implementation method.
For more information, check out: https://pennylane.ai/
"""

from .drawer import Drawer



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""A Quantum Neural Network using parameter shift to get gradients, which is
useful for sampling measurement.
"""

from __future__ import annotations

from typing import Any, Callable

import numpy as np

from .neural_network import NeuralNetwork

Tensor = Any


class QNNParamShift(NeuralNetwork):
    def __init__(
        self,
        circuit: Callable[[Tensor, Tensor], Tensor],
        num_inputs: int,
        num_weights: int,
        output_shape: int | tuple[int, ...],
        input_gradients: bool = True,
        sparse: bool = False,
    ) -> None:
        super().__init__(
            num_inputs=num_inputs,
            num_weights=num_weights,
            sparse=sparse,
            output_shape=output_shape,
            input_gradients=input_gradients,
        )

        self.circuit = circuit

    def parameter_shift_input_element(self, input, weights, i):
        input = input.copy()
        input[i] += np.pi / 2
        forwards = np.array([self.circuit(input, weights)]).reshape(-1)
        input[i] -= np.pi
        backwards = np.array([self.circuit(input, weights)]).reshape(-1)
        return 0.5 * (forwards - backwards)

    def parameter_shift_weights_element(self, input, weights, i):
        weights = weights.copy()
        weights[i] += np.pi / 2
        forwards = np.array([self.circuit(input, weights)]).reshape(-1)
        weights[i] -= np.pi
        backwards = np.array([self.circuit(input, weights)]).reshape(-1)
        return 0.5 * (forwards - backwards)

    def parameter_shift_input(self, input, weights):
        input_len = len(input)
        for i in range(input_len):
            if i == 0:
                input_grad = self.parameter_shift_input_element(input, weights, i)
                continue

            input_grad = np.append(
                input_grad,
                self.parameter_shift_input_element(input, weights, i),
            )
        input_grad = input_grad.reshape(input_len, -1).T
        return input_grad

    def parameter_shift_weights(self, input, weights):
        weights_len = len(weights)
        for i in range(weights_len):
            if i == 0:
                weights_grad = self.parameter_shift_weights_element(input, weights, i)
                continue

            weights_grad = np.append(
                weights_grad,
                self.parameter_shift_weights_element(input, weights, i),
            )
        weights_grad = weights_grad.reshape(weights_len, -1).T
        return weights_grad

    def _forward(
        self,
        inputs: Tensor,
        weights: Tensor,
    ) -> Tensor:
        """The implementation of forward calculation."""

        inputs = np.array(inputs)
        weights = np.array(weights)

        for idx, input in enumerate(inputs):
            if idx == 0:
                output_data = np.array([self.circuit(input, weights)]).reshape(-1)
                continue
            output_data = np.append(
                output_data,
                np.array([self.circuit(input, weights)]).reshape(-1),
            )

        output_data = output_data.reshape(-1, len(inputs))

        return output_data

    def _backward(
        self,
        inputs: Tensor,
        weights: Tensor,
    ) -> Tensor:
        """The implementation of backward calculation."""

        inputs = np.array(inputs)
        weights = np.array(weights)

        for idx, input in enumerate(inputs):
            if idx == 0:
                inputs_grad = self.parameter_shift_input(input, weights)
                inputs_shape = inputs_grad.shape
                continue

            inputs_grad = np.append(
                inputs_grad, self.parameter_shift_input(input, weights), axis=0
            )

        for idx, input in enumerate(inputs):
            if idx == 0:
                weights_grad = self.parameter_shift_weights(input, weights)
                weights_shape = weights_grad.shape
                continue

            weights_grad = np.append(
                weights_grad,
                self.parameter_shift_weights(input, weights),
                axis=0,
            )

        inputs_grad = inputs_grad.reshape(-1, *inputs_shape)
        weights_grad = weights_grad.reshape(-1, *weights_shape)

        if not self._input_gradients:
            return None, weights_grad

        return inputs_grad, weights_grad



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""
All the backends needed for quantum circuit simulation.
The quantum circuit is based on parsing the qcis file.
The backend can be a simulator or hardware.
The simulator is based on state vectors.
"""

from .abstract_backend import AbstractBackend
from .arclight_backend import ArclightBackend
from .autograd_backend import AutogradBackend
from .numpy_backend import NumpyBackend
from .qcis_backend import QcisBackend
from .qcis_backend2 import QcisBackend2
from .tensorflow_backend import TensorFlowBackend
from .tiany_backend import TianyBackend
from .torch_backend import TorchBackend



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""TensorFlow backend.
"""

from __future__ import annotations

import math
import os
from collections.abc import Sequence
from typing import Any

from .state_vector_simulator import StateVectorSimulator

if os.environ.get("TF_ENABLE", "0") == "1":
    try:
        import tensorflow as tf

        tf_dtype = tf.complex64
    except Exception:
        tf_dtype = None
else:
    tf_dtype = None


Tensor = Any


class TensorFlowBackend(StateVectorSimulator):
    """This method inherits the ``StateVectorSimulator``, and uses
    tensorflow to perform the mathematical calculation.
    """

    def __init__(self, dtype=None) -> None:
        if dtype is None:
            self.dtype = tf_dtype
        else:
            self.dtype = dtype

    def as_tensor(self, tensor: Tensor) -> Tensor:
        return tf.cast(tf.convert_to_tensor(tensor), self.dtype)

    def H(self) -> Tensor:
        return tf.complex(1.0 / tf.sqrt(2.0), 0.0) * tf.constant(
            [
                [1, 1],
                [1, -1],
            ],
            dtype=self.dtype,
        )

    def X(self) -> Tensor:
        return tf.constant(
            [
                [0, 1],
                [1, 0],
            ],
            dtype=self.dtype,
        )

    def Y(self) -> Tensor:
        return tf.constant(
            [
                [0j, -1j],
                [1j, 0j],
            ],
            dtype=self.dtype,
        )

    def Z(self) -> Tensor:
        return tf.constant(
            [
                [1, 0],
                [0, -1],
            ],
            dtype=self.dtype,
        )

    def S(self) -> Tensor:
        return tf.constant(
            [
                [1 + 0j, 0j],
                [0j, 1j],
            ],
            dtype=self.dtype,
        )

    def T(self) -> Tensor:
        return tf.convert_to_tensor(
            [
                [tf.exp(tf.constant(-1j * math.pi / 8)), 0],
                [0, tf.exp(tf.constant(1j * math.pi / 8))],
            ],
            dtype=self.dtype,
        )

    def SD(self) -> Tensor:
        return tf.constant(
            [
                [1 + 0j, 0j],
                [0j, -1j],
            ],
            dtype=self.dtype,
        )

    def TD(self) -> Tensor:
        return tf.convert_to_tensor(
            [
                [tf.exp(tf.constant(1j * math.pi / 8)), 0],
                [0, tf.exp(tf.constant(-1j * math.pi / 8))],
            ],
            dtype=self.dtype,
        )

    def X2M(self) -> Tensor:
        return tf.complex(tf.sqrt(2.0) / 2.0, 0.0) * tf.constant(
            [
                [1 + 0j, 1j],
                [1j, 1 + 0j],
            ],
            dtype=self.dtype,
        )

    def X2P(self) -> Tensor:
        return tf.complex(tf.sqrt(2.0) / 2.0, 0.0) * tf.constant(
            [
                [1 + 0j, -1j],
                [-1j, 1 + 0j],
            ],
            dtype=self.dtype,
        )

    def Y2M(self) -> Tensor:
        return tf.complex(tf.sqrt(2.0) / 2.0, 0.0) * tf.constant(
            [
                [1, 1],
                [-1, 1],
            ],
            dtype=self.dtype,
        )

    def Y2P(self) -> Tensor:
        return tf.complex(tf.sqrt(2.0) / 2.0, 0.0) * tf.constant(
            [
                [1, -1],
                [1, 1],
            ],
            dtype=self.dtype,
        )

    def RX(self, theta: Tensor) -> Tensor:
        theta = tf.convert_to_tensor(theta)
        theta = tf.cast(tf.math.divide(theta, 2.0), self.dtype)
        return tf.reshape(
            tf.convert_to_tensor(
                (
                    tf.cos(theta),
                    -1j * tf.sin(theta),
                    -1j * tf.sin(theta),
                    tf.cos(theta),
                ),
            ),
            (2, 2),
        )

    def RY(self, theta: Tensor) -> Tensor:
        theta = tf.convert_to_tensor(theta)
        theta = tf.cast(tf.math.divide(theta, 2.0), self.dtype)
        return tf.reshape(
            tf.convert_to_tensor(
                (
                    tf.cos(theta),
                    -1 * tf.sin(theta),
                    tf.sin(theta),
                    tf.cos(theta),
                ),
            ),
            (2, 2),
        )

    def RZ(self, theta: Tensor) -> Tensor:
        theta = tf.convert_to_tensor(theta)
        theta = tf.cast(tf.math.divide(theta, 2.0), self.dtype)
        return tf.reshape(
            tf.convert_to_tensor(
                (
                    tf.exp(-1j * theta),
                    tf.cast(tf.constant(0.0), self.dtype),
                    tf.cast(tf.constant(0.0), self.dtype),
                    tf.exp(1j * theta),
                ),
            ),
            (2, 2),
        )

    def RXY(self, theta: Tensor, phi: Tensor) -> Tensor:
        theta = tf.convert_to_tensor(theta)
        phi = tf.convert_to_tensor(phi)

        theta = tf.cast(tf.math.divide(theta, 2.0), self.dtype)
        phi = tf.cast(tf.math.divide(phi, 2.0), self.dtype)

        return tf.reshape(
            tf.convert_to_tensor(
                (
                    tf.cos(theta),
                    -1j * tf.exp(-1j * phi) * tf.sin(theta),
                    -1j * tf.exp(1j * phi) * tf.sin(theta),
                    tf.cos(theta),
                ),
            ),
            (2, 2),
        )

    def get_zero_state(self, qnum: int) -> Tensor:
        state = tf.zeros(
            tf.pow(2, qnum),
            dtype=self.dtype,
        )
        state = tf.tensor_scatter_nd_update(state, indices=[[0]], updates=[1.0 + 0j])
        return state

    def reshape(self, state: Tensor, shape: Sequence) -> Tensor:
        return tf.reshape(state, shape)

    def ravel(self, state: Tensor) -> Tensor:
        return tf.experimental.numpy.ravel(state)

    def stack(self, states: Sequence[Tensor], axis: int = 0) -> Tensor:
        return tf.stack(states, axis=axis)

    def real(self, state: Tensor) -> Tensor:
        return tf.math.real(state)

    def sum(self, state: Tensor, axis: int | None = None) -> Tensor:
        return tf.math.reduce_sum(state, axis=axis)

    def conj(self, state: Tensor) -> Tensor:
        return tf.math.conj(state)

    def sqrt(self, state: Tensor) -> Tensor:
        return tf.sqrt(state)

    def copy(self, state: Tensor) -> Tensor:
        return tf.identity(state)

    def dot(self, state1: Tensor, state2: Tensor) -> Tensor:
        state1 = self.as_tensor(state1)
        state2 = self.as_tensor(state2)
        return tf.tensordot(state1, state2, [[0], [0]])



import pytest

pyscf = pytest.importorskip("pyscf")
openfermion = pytest.importorskip("openfermion")


import numpy as np
from openfermion.linalg import get_ground_state, get_sparse_operator
from openfermion.transforms import get_fermion_operator
from pyscf import gto, scf

from isqtools.chem.utils import get_molecular_hamiltonian, parity


def test_gs():
    bond_length = 0.75

    atom = f"""
     H 0 0 0
     H 0 0 {bond_length}
    """

    basis = "sto-3g"
    charge = 0
    spin = 0
    mymol = gto.M(atom=atom, basis=basis, charge=charge, spin=spin, verbose=0)
    nao = int(mymol.nao)
    myhf = scf.RHF(mymol).run()
    molecular_hamiltonian = get_molecular_hamiltonian(myhf)
    fermion_hamiltonian = get_fermion_operator(molecular_hamiltonian)
    qubit_hamiltonian = parity(fermion_hamiltonian, nao * 2, nao)
    qubit_hamiltonian.compress()
    sparse_operator = get_sparse_operator(qubit_hamiltonian)
    energy, _ = get_ground_state(sparse_operator)
    assert np.isclose(energy, -1.1371170673457314)
    # print("Energy:", energy)



from __future__ import annotations

import datetime
import importlib.metadata

project = "isqtools"
author = "Yusheng Yang, Guolong Cui"
copyright = f"{datetime.date.today().year}, Arclight Quantum"
version = release = importlib.metadata.version("isqtools")
# version = release = "0.1.13"

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "nbsphinx",
]

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "deflist",
    "colon_fence",
]

todo_include_todos = True
napoleon_google_docstring = True
napoleon_numpy_docstring = True
copybutton_prompt_text = r">>> |\$ "
copybutton_prompt_is_regexp = True

nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc={'figure.dpi': 96}",
]
nbsphinx_prompt_width = "0"

source_suffix = [".rst", ".md"]
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]


pygments_style = "colorful"
add_module_names = False
numfig = True
numfig_format = {"table": "Table %s"}
language = "en"


autoclass_content = "both"
always_document_param_types = True

html_theme = "furo"
html_title = f"{project} {version}"
html_theme_options = {}
html_context = {"copyright": copyright}
html_last_updated_fmt = datetime.datetime.now().strftime("%b %d, %Y")
html_static_path = ["_static"]
html_css_files = ["custom.css"]

pygments_style = "default"  # Light mode friendly style
pygments_dark_style = "monokai"  # Optional dark mode style


nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
]

# inter sphinx examples
# intersphinx_mapping = {
#     "python": ("https://docs.python.org/3", None),
# }



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Abstract backend."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

Tensor = Any


class AbstractBackend(ABC):
    """Abstract backend to run quantum circuits. This is the basis for
    all simulation and hardware methods.
    """

    @abstractmethod
    def probs(self) -> Tensor:
        """This method can get all the probability distributions
        after the measurement of the specified qubit. The probability
        distribution obtained by this method is the theoretical value
        after the state vector simulation, and shots are not considered.
        Therefore, this method can only be used for circuit simulation
        and cannot be used for hardware backends.

        Returns:
            Tensor(array) distribution. Sort by the size of the binary
            representation of the quantum state.

        """

    @abstractmethod
    def sample(self) -> dict[str, int]:
        """This method can get the sampling result of the specified qubit
        measurement. Unlike the ``probs`` method, this method obtains the
        measurement results of the quantum circuit by sampling, and the
        results are related to shots. This method can be implemented using
        a simulator or the specific hardware.

        Returns:
            This method returns a dictionary, the key in the dictionary is
            the quantum state, and the value is the shots of the quantum state.

        """



import tempfile
from pathlib import Path

import pytest

from isqtools.utils import CompileTarget, IsqcError, isqc_compile, isqc_simulate

TEST_FILE_CONTENT = """
import std;

qbit q[2];

procedure main() {
    H(q[0]);
    H(q[1]);
    M(q[0]);
    M(q[1]);
}
"""


@pytest.fixture
def so_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(TEST_FILE_CONTENT)
        isqc_compile(file=str(temp_file_path), target=CompileTarget.QIR)
        output_file = temp_file_path.with_suffix(".so")
        yield output_file


def test_simulate_success(so_file):
    result, err, code = isqc_simulate(so_file)
    # print(result)
    assert isinstance(result, dict)
    assert code == 0
    assert err == ""


def test_simulate_invalid_file_format():
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".txt", delete=False
    ) as temp_file:
        temp_file.write("Invalid content")
        temp_file_path = temp_file.name

    with pytest.raises(IsqcError, match="format is not supported"):
        isqc_simulate(temp_file_path)

    Path(temp_file_path).unlink()


def test_simulate_missing_file():
    non_existent_file = Path("non_existent_file.so")
    with pytest.raises(IsqcError, match="does not exist"):
        isqc_simulate(non_existent_file)


def test_simulate_with_int_param(so_file):
    result, err, code = isqc_simulate(so_file, int_param=42)
    assert isinstance(result, dict)
    assert code == 0
    assert err == ""


def test_simulate_with_double_param(so_file):
    result, err, code = isqc_simulate(so_file, double_param=3.14)
    assert isinstance(result, dict)
    assert code == 0
    assert err == ""


def test_simulate_with_debug(so_file):
    result, err, code = isqc_simulate(so_file, debug=True)
    assert isinstance(result, dict)
    assert code == 0



import tempfile
from pathlib import Path

from isqtools.utils import CompileTarget, gen_qcis_from_so, isqc_compile

TEST_FILE_CONTENT = """
import std;

qbit q[1];

procedure main(int x[], double d[]) {
    H(q[x[0]]);
    Rx(d[0], q[x[1]]);
    M(q[x[2]]);
}
"""


def test_gen_qcis_from_so():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_isq_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_isq_file_path, "w") as temp_file:
            temp_file.write(TEST_FILE_CONTENT)
        isqc_compile(
            file=str(temp_isq_file_path),
            target=CompileTarget.QCIS,
            int_param=[0, 0, 0],  # fake data
            double_param=[0.5],
        )
        output_so_file = temp_isq_file_path.with_suffix(".so")
        gen_qcis_from_so(
            file=str(output_so_file),
            int_param=[0, 0, 0],  # real data
            double_param=[0.8],
        )
        output_qcis_file = temp_isq_file_path.with_suffix(".qcis")
        with open(output_qcis_file, "r") as f:
            assert f.read() == """H Q0\nRX Q0 0.8\nM Q0\n"""
        assert output_qcis_file.exists()



from __future__ import annotations

import json
import os
import sys
from collections import defaultdict
from datetime import datetime

from cqlib import TianYanPlatform  # 导出量子计算机SDK的支持包

from .hardware_backend import HardwareBackend


class TianyBackendError(Exception):
    """tianyan backend error"""


class TianyBackend(HardwareBackend):
    """The backend of Tianyan quantum hardware.

    token (str, optional): API token for authentication just passed when you are the first time use this backend or want to change your token. Defaults to None.

    machine_name (str, optional): Name of the machine. Defaults to full amplitude simulator .

    max_wait_time (int, optional): Maximum wait time in seconds. Defaults to 24*3600 (24 hours).

    lab_id (int | None, optional): Lab ID. Defaults to None.
    """

    def __init__(
        self,
        token: str | None = None,
        machine_name: str = "tianyan_sw",  # 默认是模拟器
        lab_id: int | None = None,
        max_wait_time: int = 24 * 3600,  # 24 hours
        **kw,
    ) -> None:

        self._token = token
        self._machine_name = machine_name
        self._max_wait_time = max_wait_time
        self._token_dir = os.path.join(self.get_homedir(), ".tianyan/")

        if token is None:
            self._token = self._load_account()
        else:
            self._token = token
            self._save_apitoken(self._token)
        if lab_id:
            self._lab_id = lab_id
        else:
            self._lab_id = f'lab.{datetime.now().strftime("%Y%m%d%H%M%S")}'

    def get_task_id(self, qcis, shots):
        self.platform = TianYanPlatform(
            login_key=self._token, machine_name=self._machine_name
        )
        lab_id = self.platform.create_lab(
            name=f"lab.{self._lab_id}", remark="test_collection"
        )
        self._query_id_single = self.platform.submit_job(
            circuit=qcis,
            lab_id=lab_id,
            exp_name=f'exp.{datetime.now().strftime("%Y%m%d%H%M%S")}',
            num_shots=shots,
        )

    def sample(
        self,
        qcis: str,
        shots: int = 100,
        **kwargs,
    ) -> dict[str, int]:
        """Call quantum hardware for sampling."""
        self.get_task_id(qcis, shots)
        return self.query_sample_result(self.query_id_single, self._max_wait_time)

    @property
    def query_id_single(self):
        return self._query_id_single

    def query_sample_result(self, query_id_single: list[str], max_wait_time: int = 1):
        try:
            measure_result = self.platform.query_experiment(
                query_id=query_id_single,
                max_wait_time=max_wait_time,
                sleep_time=0,
            )
        except Exception:
            print("Task is not done.")
            return []
        sample = self._stream_to_dict(measure_result)
        return sample

    def _load_account(self):
        file_dir = os.path.join(self._token_dir, "api")
        try:
            with open(file_dir, "r") as f:  # pylint: disable=unspecified-encoding
                data = json.load(f)
                token = data["token"]
        except FileNotFoundError as exc:
            raise TianyBackendError("Please provide the token.") from exc
        return token

    def _save_apitoken(self, token):
        data = {"token": token}
        file_path = os.path.join(self._token_dir, "api")
        if not os.path.exists(self._token_dir):
            os.mkdir(self._token_dir)
        with open(
            file_path, "w"
        ) as f:  # noqa:SCS109 # pylint: disable=unspecified-encoding
            json.dump(data, f)  # 保存为字典

    def _stream_to_dict(self, measure_result):
        answer = defaultdict(int)
        # defaultdict is used for data processing.
        # print(1112, measure_result[0]["probability"])
        for t in measure_result[0]["resultStatus"][1:]:
            x = "".join([str(v) for v in t])
            answer[x] += 1
        res = dict(answer)
        return res

    @staticmethod
    def get_homedir():
        if sys.platform == "win32":
            return os.environ["USERPROFILE"]
        if sys.platform in ["darwin", "linux"]:
            return os.environ["HOME"]
        raise ValueError(
            f"unsupported platform:{sys.platform}. "
            f"You may raise a request issue on github."
        )



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""State vector simulation backend."""

from __future__ import annotations

from collections import Counter, defaultdict
from collections.abc import Iterable, Sequence
from typing import Any

import numpy as np

from .abstract_backend import AbstractBackend

Tensor = Any

gate_list: list[str] = [
    "H",
    "X",
    "Y",
    "Z",
    "S",
    "T",
    "RZ",
    "RX",
    "RY",
    "SD",
    "TD",
    "X2M",
    "X2P",
    "Y2M",
    "Y2P",
    "CZ",
    "CY",
    "CX",
    "CNOT",
    "M",
    "RXY",
]


class StateVectorSimulatorError(Exception):
    """state vector simulation error"""


class StateVectorSimulator(AbstractBackend):
    """Simulate quantum circuits using state vector methods. Abstract the
    mathematical operations required for the simulation process, and the
    specific implementation methods need to be implemented in its subclasses.
    """

    def as_tensor(self, Tensor) -> Tensor:
        raise NotImplementedError

    def H(self) -> Tensor:
        raise NotImplementedError

    def X(self) -> Tensor:
        raise NotImplementedError

    def Y(self) -> Tensor:
        raise NotImplementedError

    def Z(self) -> Tensor:
        raise NotImplementedError

    def S(self) -> Tensor:
        raise NotImplementedError

    def T(self) -> Tensor:
        raise NotImplementedError

    def SD(self) -> Tensor:
        raise NotImplementedError

    def TD(self) -> Tensor:
        raise NotImplementedError

    def X2M(self) -> Tensor:
        raise NotImplementedError

    def X2P(self) -> Tensor:
        raise NotImplementedError

    def Y2M(self) -> Tensor:
        raise NotImplementedError

    def Y2P(self) -> Tensor:
        raise NotImplementedError

    def RX(self, theta: Tensor) -> Tensor:
        raise NotImplementedError

    def RY(self, theta: Tensor) -> Tensor:
        raise NotImplementedError

    def RZ(self, theta: Tensor) -> Tensor:
        raise NotImplementedError

    def RXY(self, theta: Tensor, phi: Tensor) -> Tensor:
        raise NotImplementedError

    def reshape(self, state: Tensor, shape: Sequence) -> Tensor:
        raise NotImplementedError

    def ravel(self, state: Tensor) -> Tensor:
        raise NotImplementedError

    def stack(self, states: Sequence[Tensor], axis: int = 0) -> Tensor:
        raise NotImplementedError

    def real(self, state: Tensor) -> Tensor:
        raise NotImplementedError

    def sum(self, state: Tensor, axis: int | None = None) -> Tensor:
        raise NotImplementedError

    def conj(self, state: Tensor) -> Tensor:
        raise NotImplementedError

    def sqrt(self, state: Tensor) -> Tensor:
        raise NotImplementedError

    def copy(self, state: Tensor) -> Tensor:
        raise NotImplementedError

    def dot(self, state1: Tensor, state2: Tensor) -> Tensor:
        raise NotImplementedError

    def get_zero_state(self, qnum: int) -> Tensor:
        raise NotImplementedError

    def reshape_single(
        self,
        state: Tensor,
        qnum: int,
        target: int,
    ) -> Tensor:
        """Before the single-qubit operation is performed, reshape the state
        vector into the required dimension for matrix multiplication.

        Args:
            state: Initial state vector.
            qnum: Number of qubits.
            target: The target qubit corresponding to the action of a
                    single-qubit operation.

        Returns:
            Reshaped state vector.

        """

        shape = (1 << target, 2, 1 << (qnum - target - 1))
        return self.reshape(state, shape)

    def reshape_double(
        self,
        state: Tensor,
        qnum: int,
        target1: int,
        target2: int,
    ) -> Tensor:
        """Before the double-qubit operation is performed, reshape the state
        vector into the required dimension for matrix multiplication.

        Args:
            state: Initial state vector.
            qnum: Number of qubits.
            target1: The first target qubit corresponding to the action of a
                     double-qubit operation.
            target2: The second target qubit corresponding to the action of a
                     double-qubit operation.

        Returns:
            Reshaped state vector.

        """
        shape = (
            1 << target1,
            2,
            1 << (target2 - target1 - 1),
            2,
            1 << (qnum - target2 - 1),
        )
        return self.reshape(state, shape)

    def single_gate(
        self,
        state: Tensor,
        gate: str,
        qnum: int,
        target: int,
    ) -> Tensor:
        """Using matrix multiplication, single-qubit gates act on quantum state
        vectors.

        Args:
            state: Initial state vector before operation.
            gate: Specific gates acting on the qubit.
            qnum: Number of qubits.
            target: The target qubit corresponding to the action.

        Returns:
            Final state vector.

        """
        state = self.reshape_single(state, qnum, target)
        if gate in gate_list:
            state = getattr(self, gate)() @ state[:,]
        else:
            raise StateVectorSimulatorError("non-existent quantum gate")

        return self.ravel(state)

    def single_rotate_gate(
        self,
        state: Tensor,
        gate: str,
        qnum: int,
        target: int,
        theta: Tensor,
    ) -> Tensor:
        """Using matrix multiplication, single-qubit-rotate gates act on
        quantum state vectors.

        Args:
            state: Initial state vector before operation.
            gate: Specific gates acting on the qubit.
            qnum: Number of qubits.
            target: The target qubit corresponding to the action.
            theta: Angle of rotation. Extra dimensions after the parameter
                   needed will be discarded.

        Returns:
            Final state vector.

        """
        state = self.reshape_single(state, qnum, target)
        if gate in ["RX", "RY", "RZ"]:
            state = getattr(self, gate)(theta[0]) @ state[:,]
        elif gate == "RXY":
            state = self.RXY(theta[0], theta[1]) @ state[:,]
        return self.ravel(state)

    def multi_gate(
        self,
        state: Tensor,
        gate: str,
        qnum: int,
        ctrl: int,
        target: int,
    ) -> Tensor:
        """Using matrix multiplication, multi-qubit gates act on quantum state
        vectors.

        Args:
            state: Initial state vector before operation.
            gate: Specific gates acting on the qubit.
            qnum: Number of qubits.
            ctrl: The control qubit corresponding to the action.
            target: The target qubit corresponding to the action.

        Returns:
            Final state vector.

        """

        if ctrl < target:
            state = self.reshape_double(state, qnum, ctrl, target)
            if gate in ["CX", "CNOT"]:
                a, b = state[:, 1, :, 0, :], state[:, 1, :, 1, :]
                u = self.stack([b, a], axis=2)
                state = self.stack([state[:, 0, :, :, :], u], axis=1)
            elif gate == "CY":
                a, b = 1j * state[:, 1, :, 0, :], -1j * state[:, 1, :, 1, :]
                u = self.stack([b, a], axis=2)
                state = self.stack([state[:, 0, :, :, :], u], axis=1)
            elif gate == "CZ":
                a, b = state[:, 0, :, 1, :], -1 * state[:, 1, :, 1, :]
                u = self.stack([a, b], axis=1)
                state = self.stack([state[:, :, :, 0, :], u], axis=3)
            return self.ravel(state)
        else:
            state = self.reshape_double(state, qnum, target, ctrl)
            if gate in ["CX", "CNOT"]:
                a, b = state[:, 0, :, 1, :], state[:, 1, :, 1, :]
                u = self.stack([b, a], axis=1)
                state = self.stack([state[:, :, :, 0, :], u], axis=3)
            elif gate == "CY":
                a, b = 1j * state[:, 0, :, 1, :], -1j * state[:, 1, :, 1, :]
                u = self.stack([b, a], axis=1)
                state = self.stack([state[:, :, :, 0, :], u], axis=3)
            elif gate == "CZ":
                a, b = state[:, 0, :, 1, :], -1 * state[:, 1, :, 1, :]
                u = self.stack([a, b], axis=1)
                state = self.stack([state[:, :, :, 0, :], u], axis=3)
            return self.ravel(state)

    def swap(
        self,
        state: Tensor,
        qnum: int,
        q1: int,
        q2: int,
    ) -> Tensor:
        """The state vector is swapped, and the sequence number of the qubit
        is exchanged.

        Args:
            state: State vector before swap.
            qnum: Number of qubits.
            q1, q2: Indexed of qubits to exchange.

        Returns:
            State vector after swap.

        """
        q1, q2 = min(q1, q2), max(q1, q2)
        state = self.reshape_double(state, qnum, q1, q2)
        a, b, c, d = (
            state[:, 0, :, 0, :],
            state[:, 0, :, 1, :],
            state[:, 1, :, 0, :],
            state[:, 1, :, 1, :],
        )
        u = self.stack([a, c], axis=2)
        v = self.stack([b, d], axis=2)
        state = self.stack([u, v], axis=1)
        return self.ravel(state)

    def shift(
        self,
        state: Tensor,
        qnum: int,
        mq: Sequence[int],
    ) -> Tensor:
        """Qubits are swapped multiple times for ``probs``.

        Args:
            state: State vector before shift.
            qnum: Number of qubits.
            mq: The number of qubits to be measured.

        Returns:
            State vector after shift.

        """

        qidx = {}
        idxq = {}
        for i in range(qnum):
            qidx[i] = i
            idxq[i] = i

        for i, m in enumerate(mq):
            if qidx[m] == i:
                continue
            state = self.swap(state, qnum, i, qidx[m])
            q = idxq[i]
            qidx[q] = qidx[m]
            qidx[m] = i
            idxq[i] = m
            idxq[qidx[q]] = q
        return state

    @staticmethod
    def check(line_data: Iterable) -> tuple[int, dict]:
        """Traverse the entire qcis to get circuit information for simulation.

        Args:
            line_data: Each line of the qcis file.

        Returns:
            qnum: The number of qubits required for the circuit.
            qdic: The number corresponding to the qubit.

        Raises:
            StateVectorSimulatorError: Check the rationality of the qcis file.

        """

        qdic = {}
        qnum = 0
        for idx, line in enumerate(line_data):
            line = line.strip()
            if not line:
                continue
            strArr = line.split(" ")
            if strArr[0] not in gate_list:
                raise StateVectorSimulatorError(
                    f"simulate error: in line {idx}, gate error"
                )
            if len(strArr) < 2 or len(strArr) > 4:
                raise StateVectorSimulatorError(
                    f"simulate error: in line {idx}, qbit number error"
                )
            if strArr[1][0] != "Q" or not strArr[1][1:].isdigit():
                raise StateVectorSimulatorError(
                    f"simulate error: in line {idx}, qbit syntax error"
                )

            if strArr[1] not in qdic:
                qdic[strArr[1]] = qnum
                qnum += 1

            if strArr[0] in ["CZ", "CY", "CX", "CNOT"]:
                if len(strArr) != 3:
                    raise StateVectorSimulatorError(
                        f"simulate error: in line {idx}, qbit number error"
                    )

                if strArr[2][0] != "Q" or not strArr[2][1:].isdigit():
                    raise StateVectorSimulatorError(
                        f"simulate error: in line {idx}, qbit syntax error"
                    )

                if strArr[2] not in qdic:
                    qdic[strArr[2]] = qnum
                    qnum += 1
            if strArr[0] in ["RX", "RY", "RZ"]:
                if len(strArr) != 3:
                    raise StateVectorSimulatorError(
                        f"simulate error: in line {idx}, qbit number error"
                    )
            if strArr[0] == "RXY":
                if len(strArr) != 4:
                    raise StateVectorSimulatorError(
                        f"simulate error: in line {idx}, qbit number error"
                    )

        if qnum > 30:
            raise StateVectorSimulatorError(
                f"simulate error: qbit number `{qnum}` is too large, "
                "can not simulate."
            )

        if qnum > 22:
            print(
                f"Warning: current number of qubits is {qnum}, "
                "which may require a lot of memory."
            )

        return qnum, qdic

    def getstate(
        self,
        line_data: Iterable,
        qnum: int,
        qdic: dict,
        **kwargs,
    ) -> tuple[Tensor, list]:
        """Initialize the state vector, traverse the qcis file, apply all gate
        operations to the state vector, obtain the state after the action,
        and obtain the number of the measurement qubit.

        Args:
            line_data: Each line of the qcis file.
            qnum: The number of qubits.
            qdic: The number corresponding to the qubit.
            **kwargs: The parameters of the rotation gates.

        Returns:
            state: Final state vector.
            mq: The sequence number of the qubit to be measured.

        """

        state = self.get_zero_state(qnum)
        mq = []
        for line in line_data:
            line = line.strip()
            if not line:
                continue
            strArr = line.split(" ")
            qid1 = qdic[strArr[1]]
            if strArr[0] == "M":
                mq.append(qid1)
            else:
                if strArr[0] in ["CZ", "CY", "CX", "CNOT"]:
                    qid2 = qdic[strArr[2]]
                    state = self.multi_gate(state, strArr[0], qnum, qid1, qid2)
                elif strArr[0] in ["RX", "RY", "RZ", "RXY"]:
                    theta = []
                    for v in strArr[2:]:
                        theta.append(eval(v, kwargs))
                    state = self.single_rotate_gate(state, strArr[0], qnum, qid1, theta)
                else:
                    state = self.single_gate(state, strArr[0], qnum, qid1)
        return state, mq

    def sample(
        self,
        data: str,
        shots: int = 100,
        rng_seed: int | None = None,
        **kwargs,
    ) -> dict[str, int]:
        """This method can get the sampling result of the specified qubit
        measurement. Unlike the ``probs`` method, this method obtains the
        measurement results of the quantum circuit by sampling, and the results
        are related to shots.

        Returns:
            This method returns a dictionary, the key in the dictionary is
            the quantum state, and the value is the shots of the quantum state.

        """
        line_data = data.split("\n")
        qnum, qdic = self.check(line_data)
        state, mq = self.getstate(line_data, qnum, qdic, **kwargs)
        state = self.shift(state, qnum, mq)
        state = self.conj(state) * state
        mq_len = len(mq)
        state = self.reshape(state, [1 << mq_len, 1 << (qnum - mq_len)])
        p = self.real(self.sum(state, axis=1))
        p_norm = p / sum(p)
        if not isinstance(p_norm, np.ndarray):
            p_norm = p_norm.cpu().numpy()
        if rng_seed is not None:
            np.random.seed(int(rng_seed))
        r = np.random.choice(1 << mq_len, shots, p=p_norm)
        return {bin(k)[2:].zfill(mq_len): v for k, v in Counter(r).items()}

    def probs(
        self,
        data: str,
        **kwargs,
    ) -> Tensor:
        """This method can get all the probability distributions after the
        measurement of the specified qubit. The probability distribution
        obtained by this method is the theoretical value after the state vector
        simulation, and shots are not considered.

        Args:
            data: Qcis data.

        Returns:
            Tensor(array) distribution. Sort by the size of the binary.

        """
        line_data = data.split("\n")
        qnum, qdic = self.check(line_data)
        state, mq = self.getstate(line_data, qnum, qdic, **kwargs)
        state = self.shift(state, qnum, mq)
        state = self.conj(state) * state
        mq_len = len(mq)
        state = self.reshape(state, [1 << mq_len, 1 << (qnum - mq_len)])
        return self.real(self.sum(state, axis=1))

    def state(
        self,
        data: str,
        **kwargs,
    ) -> Tensor:
        """This method can get the quantum state after operation.

        Args:
            data: Qcis data.

        Returns:
            State as tensor(array).

        """

        line_data = data.split("\n")
        qnum, qdic = self.check(line_data)
        state, _ = self.getstate(line_data, qnum, qdic, **kwargs)
        return state



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Arclight quantum cloud backend.
For more information, check out: http://qcloud.arclightquantum.com/#/home
"""

from __future__ import annotations

import json
import os
import time

from .hardware_backend import HardwareBackend, ISQTask, TaskState, TaskType
from .qcis_backend import get_rand_str, load_params, split_rotation_gates

try:
    import requests
except ModuleNotFoundError:
    pass


class ArclightBackend(HardwareBackend):
    """The backend of arclighth quantum cloud."""

    def __init__(
        self,
        token: str | None = None,
        version: str | None = None,
        # max_wait_time: int = 60,
        sleep_time: int = 3,
        run_time: int | None = None,
        mapping: bool = False,
    ) -> None:
        if token is None:
            token_path = os.path.expanduser("~/.isq/token")
            with open(token_path, "r") as token_file:
                token = token_file.read()
        self.headers = {"Authorization": token}
        self.host = "https://qcloud.arclightquantum.com"

        if version is None:
            self.version = ""
        else:
            self.version = str(version)

        # self._max_wait_time = max_wait_time
        self.sleep_time = sleep_time
        self.run_time = run_time
        self.mapping = mapping
        if self.mapping:
            print("Mapping is not implemented yet.")
        # TODO: Implement mapping

    def _post_circuit(
        self,
        circuit: str,
        shots: int,
    ) -> dict:
        if self.version:
            version_posted = f"{self.version}_{get_rand_str()}"
        else:
            version_posted = get_rand_str()

        post_url = f"{self.host}/qcloud/task/submit/job"
        param = {
            "circuit": circuit,
            "shots": shots,
            "version": version_posted,
        }

        post_response = requests.post(
            post_url,
            json=param,
            headers=self.headers,
        )

        return self._response_process(post_response)

    def _get_result(self, id: str) -> dict:
        get_url = f"{self.host}/qcloud/task/info/{id}"
        get_response = requests.get(get_url, headers=self.headers)
        return self._response_process(get_response)

    @staticmethod
    def _response_process(
        response: requests.models.Response,
    ) -> dict:
        status_code = response.status_code
        if status_code == 200:
            response_dict = json.loads(response.text)
            if response_dict.get("code") == 0:
                return response_dict
            elif response_dict.get("code") == -1:
                print(response_dict.get("msg"))
                return {}
        # FIX: not print every `Error code`
        else:
            print(f"Error code: {status_code}.")
            return {}

    def sample(
        self,
        data: str,
        shots: int = 100,
        **kwargs,
    ) -> dict[str, int]:
        """Call quantum hardware for sampling.

        Args:
            data: QCIS strings, This represents the information of the quantum
                  circuit.
            shots: Shots numbers.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            This method returns a dictionary, the key in the dictionary is
            the quantum state, and the value is the shots of the quantum state.

        """

        qcis_split_rotation_gates = split_rotation_gates(
            load_params(data, **kwargs),
        )

        post_dict = self._post_circuit(
            circuit=qcis_split_rotation_gates,
            shots=shots,
        )

        if post_dict:
            task = ISQTask(
                post_dict.get("data"),
                TaskType.ARCLIGHT,
                TaskState.WAIT,
                self,
            )
        else:
            task = ISQTask(
                0,
                TaskType.ARCLIGHT,
                TaskState.FAIL,
                self,
            )

        if self.run_time is not None:
            start_time = time.time()

        while task.state == TaskState.WAIT:
            print("request_test")
            task.result()
            time.sleep(self.sleep_time)
            # Wait for a period of time before doing an http request
            if self.run_time is not None:
                if time.time() > start_time + self.run_time:
                    break
        return task.result()



import pytest

autograd = pytest.importorskip("autograd")


import tempfile
from pathlib import Path

import numpy as np
from autograd import grad

from isqtools import IsqCircuit
from isqtools.backend import AutogradBackend

GRAD_ISQ = """
import std;


param params[];
qbit q[1];

procedure main() {
    Ry(params[0], q[0]);
    Rx(params[1], q[0]);
    M(q[0]);
}
"""


def test_autograd():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(GRAD_ISQ)

        autograd_backend = AutogradBackend()
        params = np.array([0.9, 1.2])
        qc = IsqCircuit(
            file=str(temp_file_path),
            backend=autograd_backend,
            sample=False,
        )

        def circuit(params):
            results = qc.measure(params=params)
            return results[0]

        assert circuit(params) == pytest.approx(0.6126225961314372)
        grad_circuit1 = grad(circuit, [0])
        np.testing.assert_allclose(
            grad_circuit1(params)[0], np.array([-0.14192229, -0.28968239])
        )



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Python quantum circuit class.
We recommend using ``isqc`` to compile and generate the Qics instruction set to
build quantum circuits. However, to guarantee support for building circuits
via python, we built ``StrCircuit``. This class constructs quantum circuits in
the form of strings. This method cannot use the advanced features of ``isqc``
and is not recommended.
"""

from __future__ import annotations

import itertools
import re
from collections.abc import Sequence
from typing import Any

from isqtools.backend import AbstractBackend

from .abstract_circuit import AbstractCircuit

Tensor = Any


class StrCircuitError(Exception):
    """Python string quantum circuits error."""


class StrCircuit(AbstractCircuit):
    """Python string quantum circuit class.

    Attributes:
        num_qubits: Number of qubits simulated.
        backend: The backend on which the task runs can be a simulator or
                 hardware.
        sample: Whether to enable sampling mode. When the backend is a
                simulator, you can ``sample`` or ``probs``. If the backend is
                hardware, you can only enable the sampling mode.
        shots: When the sampling mode is enabled, this parameter can specify
               the shots number.
        qcis_list: Use list to record each Qcis instruction set.

    """

    def __init__(
        self,
        num_qubits: int | None,
        backend: AbstractBackend,
        sample: bool = False,
        shots: int = 100,
    ) -> None:
        super().__init__(
            backend=backend,
            sample=sample,
            shots=shots,
        )

        self.qcis_list = []  # to record qcis
        self._qbit_init(num_qubits)
        self.num_qubits = num_qubits

    @property
    def qcis(self) -> str:
        """Join ``qcis_list`` to get qcis strings."""
        return "\n".join(self.qcis_list)

    def __len__(self) -> int:
        return len(self.qcis_list)

    def __getitem__(self, key: int) -> str:
        return self.qcis_list[key]

    def __iter__(self):
        return (item for item in self.qcis_list)

    def _qbit_init(
        self,
        num_qubits: int,
    ) -> None:
        """Qubits initialization.

        Raises:
            StrCircuitError: Guarantee that the number of qubits is not
            negative.

        """
        if num_qubits < 0:
            raise StrCircuitError(f"Number of inputs cannot be negative: {num_qubits}!")

    def _one_qubit_op(
        self,
        qubit_idx: int,
        operation: str,
    ) -> None:
        """General one qubit operations."""
        self.qcis_list.append(f"{operation} Q{qubit_idx}")

    def h(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "H"
        self._one_qubit_op(qubit_idx, operation)

    def x(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "X"
        self._one_qubit_op(qubit_idx, operation)

    def y(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "Y"
        self._one_qubit_op(qubit_idx, operation)

    def z(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "Z"
        self._one_qubit_op(qubit_idx, operation)

    def s(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "S"
        self._one_qubit_op(qubit_idx, operation)

    def sd(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "SD"
        self._one_qubit_op(qubit_idx, operation)

    def t(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "T"
        self._one_qubit_op(qubit_idx, operation)

    def td(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "TD"
        self._one_qubit_op(qubit_idx, operation)

    def x2p(
        self,
        qubit_idx: int,
    ) -> None:
        """X rotate pi/2"""
        operation = "X2P"
        self._one_qubit_op(qubit_idx, operation)

    def x2m(
        self,
        qubit_idx: int,
    ) -> None:
        """X rotate -pi/2"""
        operation = "X2M"
        self._one_qubit_op(qubit_idx, operation)

    def y2p(
        self,
        qubit_idx: int,
    ) -> None:
        """Y rotate pi/2"""
        operation = "Y2P"
        self._one_qubit_op(qubit_idx, operation)

    def y2m(
        self,
        qubit_idx: int,
    ) -> None:
        """Y rotate -pi/2"""
        operation = "Y2M"
        self._one_qubit_op(qubit_idx, operation)

    def m(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "M"
        self._one_qubit_op(qubit_idx, operation)

    def H(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "H"
        self._one_qubit_op(qubit_idx, operation)

    def X(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "X"
        self._one_qubit_op(qubit_idx, operation)

    def Y(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "Y"
        self._one_qubit_op(qubit_idx, operation)

    def Z(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "Z"
        self._one_qubit_op(qubit_idx, operation)

    def S(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "S"
        self._one_qubit_op(qubit_idx, operation)

    def SD(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "SD"
        self._one_qubit_op(qubit_idx, operation)

    def T(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "T"
        self._one_qubit_op(qubit_idx, operation)

    def TD(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "TD"
        self._one_qubit_op(qubit_idx, operation)

    def X2P(
        self,
        qubit_idx: int,
    ) -> None:
        """X rotate pi/2"""
        operation = "X2P"
        self._one_qubit_op(qubit_idx, operation)

    def X2M(
        self,
        qubit_idx: int,
    ) -> None:
        """X rotate -pi/2"""
        operation = "X2M"
        self._one_qubit_op(qubit_idx, operation)

    def Y2P(
        self,
        qubit_idx: int,
    ) -> None:
        """Y rotate pi/2"""
        operation = "Y2P"
        self._one_qubit_op(qubit_idx, operation)

    def Y2M(
        self,
        qubit_idx: int,
    ) -> None:
        """Y rotate -pi/2"""
        operation = "Y2M"
        self._one_qubit_op(qubit_idx, operation)

    def M(
        self,
        qubit_idx: int,
    ) -> None:
        operation = "M"
        self._one_qubit_op(qubit_idx, operation)

    def _two_qubits_op(
        self,
        qubit_control: int,
        qubit_target: int,
        operation: str,
    ) -> None:
        """
        General two qubit operations.
        """
        self.qcis_list.append(f"{operation} Q{qubit_control} Q{qubit_target}")

    def cnot(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        # FIX: not supported in hardware, just for simulation
        operation = "CX"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def cx(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        # FIX: not supported in hardware, just for simulation
        operation = "CX"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def cy(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        # FIX: not supported in hardware, just for simulation
        operation = "CY"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def cz(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        operation = "CZ"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def CNOT(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        # FIX: not supported in hardware, just for simulation
        operation = "CX"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def CX(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        # FIX: not supported in hardware, just for simulation
        operation = "CX"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def CY(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        # FIX: not supported in hardware, just for simulation
        operation = "CY"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def CZ(
        self,
        qubit_control: int,
        qubit_target: int,
    ) -> None:
        operation = "CZ"
        self._two_qubits_op(qubit_control, qubit_target, operation)

    def _rotate_gate(
        self,
        param: float | str,
        qubit_idx: int,
        operation: str,
    ) -> None:
        """
        General rotion operations.
        """
        self.qcis_list.append(f"{operation} Q{qubit_idx} {param}")

    def rx(
        self,
        param: float | str,
        qubit_idx: int,
    ) -> None:
        operation = "RX"
        self._rotate_gate(param, qubit_idx, operation)

    def ry(
        self,
        param: float | str,
        qubit_idx: int,
    ) -> None:
        operation = "RY"
        self._rotate_gate(param, qubit_idx, operation)

    def rz(
        self,
        param: float | str,
        qubit_idx: int,
    ) -> None:
        operation = "RZ"
        self._rotate_gate(param, qubit_idx, operation)

    def RX(
        self,
        param: float | str,
        qubit_idx: int,
    ) -> None:
        operation = "RX"
        self._rotate_gate(param, qubit_idx, operation)

    def RY(
        self,
        param: float | str,
        qubit_idx: int,
    ) -> None:
        operation = "RY"
        self._rotate_gate(param, qubit_idx, operation)

    def RZ(
        self,
        param: float | str,
        qubit_idx: int,
    ) -> None:
        operation = "RZ"
        self._rotate_gate(param, qubit_idx, operation)

    def rxy(
        self,
        phi: float | str,
        theta: float | str,
        qubit_idx: int,
    ) -> None:
        self.qcis_list.append(f"RXY Q{qubit_idx} {phi} {theta}")

    RXY = rxy

    def mqbit(
        self,
        *qubit_idx: Sequence[int] | int,
    ) -> None:
        """Specifies the object to be measured."""

        for idx in qubit_idx:
            self.qcis_list.append(f"M Q{idx}")

    def measure(
        self,
        **kw,
    ):
        """Forward calculation."""
        # TODO: check qubit_num
        if not self.sample:
            measure_result = self.backend.probs(self.qcis, **kw)
        else:
            measure_result = self.backend.sample(self.qcis, self.shots, **kw)
        return measure_result

    def pauli(
        self,
        gates: str | tuple,
        format: str = "str",
    ) -> None:
        """Pauli measurement."""

        if format == "str":
            if not isinstance(gates, str):
                raise TypeError(f"Strings are needs but get f{type(gates)}")
            gate_op = re.findall("[a-zA-Z]+", gates)
            gate_idx = [int(idx) for idx in re.findall("[0-9]+", gates)]
            if len(gate_op) != len(gate_idx):
                raise ValueError("Illegal pauli measurement.")
            gates = zip(gate_idx, gate_op)

        elif format == "openfermion":
            pass

        else:
            raise StrCircuitError(f"Unsupported format {format!s}")

        for gate in gates:
            if gate[1].upper() == "X":
                self._measure_recording.append(
                    f"H Q{gate[0]}",
                )
                self._measure_recording.append(
                    f"M Q{gate[0]}",
                )
            elif gate[1].upper() == "Y":
                self._measure_recording.append(
                    f"X2P Q{gate[0]}",
                )
                self._measure_recording.append(
                    f"M Q{gate[0]}",
                )
            elif gate[1].upper() == "Z":
                self._measure_recording.append(
                    f"M Q{gate[0]}",
                )
            else:
                raise ValueError("Please input correct Pauli gates.")

    def pauli_measure(self, **kw) -> Tensor:
        """Pauli measurement."""
        # TODO: check qubit_num

        if self.sample:
            measure_result_dict = self.measure(**kw)
            result = 0
            for res_index, frequency in measure_result_dict.items():
                parity = (-1) ** (res_index.count("1") % 2)
                # e.g. {"011": 222}, to count "1"
                result += parity * frequency / self.shots
            return result
        else:
            measure_result_array = self.measure(**kw)
            parity = [
                (-1) ** (str(bin(int(index))).count("1") % 2)
                for index in range(len(measure_result_array))
            ]
            return self.backend.dot(
                measure_result_array,
                self.backend.as_tensor(parity),
            )

    def chain(
        self,
        gate_name: str,
        qubit_idx: Sequence[int] | None = None,
    ) -> None:
        """Act on each qubit and the next in turn"""
        if qubit_idx is None:
            qubit_idx = list(range(self.num_qubits - 1))
        operation = getattr(self, gate_name)
        for idx in qubit_idx:
            operation(idx, idx + 1)

    def ring(
        self,
        gate_name: str,
        num_qubits: int | None = None,
    ) -> None:
        """Act on all qubits via a ring"""
        if num_qubits is None:
            num_qubits = self.num_qubits
        if num_qubits < 2:
            raise ValueError(
                f"Number of qubits have to be larger than 1 but given " f"{num_qubits}"
            )
        operation = getattr(self, gate_name)
        for num_qubit in range(num_qubits - 1):
            operation(num_qubit, num_qubit + 1)
        operation(num_qubits - 1, 0)

    def single(
        self,
        gate_name: str,
        qubit_idx: Sequence[int] | None = None,
    ) -> None:
        """Act on each qubit in turn"""
        if qubit_idx is None:
            qubit_idx = list(range(self.num_qubits))
        operation = getattr(self, gate_name)
        for idx in qubit_idx:
            # self._one_qubit_op(idx, gate_name)
            operation(idx)

    def perm(
        self,
        gate_name: str,
        qubit_idx: Sequence[int] | None = None,
    ) -> None:
        """Permutation of all."""
        if qubit_idx is None:
            qubit_idx = list(range(self.num_qubits))
        operation = getattr(self, gate_name)
        for idx in itertools.permutations(qubit_idx, 2):
            operation(*idx)

    def comb(
        self,
        gate_name: str,
        qubit_idx: Sequence[int] | None = None,
    ) -> None:
        """combinations of all."""
        if qubit_idx is None:
            qubit_idx = list(range(self.num_qubits))
        operation = getattr(self, gate_name)
        for idx in itertools.combinations(qubit_idx, 2):
            operation(*idx)

    def param(
        self,
        gate_name: str,
        qubit_idx: Sequence[int] | None = None,
        param_name: str = "x",
        param_idx: Sequence[int] | None = None,
    ) -> None:
        """Quickly build parameterized circuits"""
        if qubit_idx is None:
            qubit_idx = list(range(self.num_qubits))
        if param_idx is None:
            param_idx = qubit_idx
        operation = getattr(self, gate_name)
        for pidx, qidx in zip(param_idx, qubit_idx):
            operation(f"{param_name}[{pidx}]", qidx)



import pytest

pyscf = pytest.importorskip("pyscf")


import numpy as np
from pyscf import cc, gto, scf


def test_model():
    bond_length = 0.75

    atom = f"""
     H 0 0 0
     H 0 0 {bond_length}
    """

    basis = "sto-3g"
    charge = 0
    spin = 0

    mymol = gto.M(atom=atom, basis=basis, charge=charge, spin=spin, verbose=0)

    nao = int(mymol.nao)
    assert nao == 2
    # print("Spin orbs:", nao)

    # Hartree-Fock calculation
    myhf = scf.RHF(mymol).run()
    print(type(myhf.e_tot))
    assert np.isclose(myhf.e_tot, -1.11615145)
    # print(f"HF energy by pyscf: {myhf.e_tot:.8f}")

    # CCSD(T) calculation
    # mycc = cc.CCSD(myhf).run()
    # et = mycc.ccsd_t()
    # assert np.isclose(mycc.e_tot + et, -1.1371172451631748)
    # print(f"CCSD(T) energy: {mycc.e_tot + et:.8f}")



import pytest

from isqtools.utils import _deal_params


@pytest.mark.parametrize(
    "int_param, double_param, expected_output",
    [
        (3, 0.5, ("-i 3", "-d 0.5")),
        ([1, 2], [0.1, 0.2], ("-i 1 -i 2", "-d 0.1 -d 0.2")),
        (None, None, ("", "")),
        ([3], None, ("-i 3", "")),
        (None, [0.1], ("", "-d 0.1")),
        ([], [], ("", "")),
        ([42], 3.14, ("-i 42", "-d 3.14")),
    ],
)
def test_deal_params(int_param, double_param, expected_output):
    assert _deal_params(int_param, double_param) == expected_output



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""PyTorch backend.
PyTorch is a machine learning framework based on the Torch library, used for
applications such as computer vision and natural language processing,
originally developed by Meta AI and now part of the Linux Foundation umbrella.
It is free and open-source software released under the modified BSD license.
For more information, check out: https://pytorch.org/
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from .state_vector_simulator import StateVectorSimulator

try:
    import torch

    torch_dtype = torch.complex64
except Exception:
    # pytorch is an optional backend
    torch_dtype = None


Tensor = Any


class TorchBackend(StateVectorSimulator):
    """This method inherits the ``StateVectorSimulator``, and uses
    pytorch to perform the mathematical calculation.
    """

    def __init__(self, dtype=None) -> None:
        if dtype is None:
            self.dtype = torch_dtype
        else:
            self.dtype = dtype

    def as_tensor(self, tensor: Tensor) -> Tensor:
        return torch.as_tensor(tensor)

    def H(self) -> Tensor:
        return (
            1
            / torch.sqrt(torch.tensor(2))
            * torch.tensor(
                [
                    [1, 1],
                    [1, -1],
                ],
                dtype=self.dtype,
            )
        )

    def X(self) -> Tensor:
        return torch.tensor(
            [
                [0, 1],
                [1, 0],
            ],
            dtype=self.dtype,
        )

    def Y(self) -> Tensor:
        return torch.tensor(
            [
                [0, -1j],
                [1j, 0],
            ],
            dtype=self.dtype,
        )

    def Z(self) -> Tensor:
        return torch.tensor(
            [
                [1, 0],
                [0, -1],
            ],
            dtype=self.dtype,
        )

    def S(self) -> Tensor:
        return torch.tensor(
            [
                [1, 0],
                [0, 1j],
            ],
            dtype=self.dtype,
        )

    def T(self) -> Tensor:
        return torch.tensor(
            [
                [torch.exp(torch.tensor(-1j * torch.pi / 8)), 0],
                [0, torch.exp(torch.tensor(1j * torch.pi / 8))],
            ],
            dtype=self.dtype,
        )

    def SD(self) -> Tensor:
        return torch.tensor(
            [
                [1, 0],
                [0, -1j],
            ],
            dtype=self.dtype,
        )

    def TD(self) -> Tensor:
        return torch.tensor(
            [
                [torch.exp(torch.tensor(1j * torch.pi / 8)), 0],
                [0, torch.exp(torch.tensor(-1j * torch.pi / 8))],
            ],
            dtype=self.dtype,
        )

    def X2M(self) -> Tensor:
        return (
            torch.sqrt(torch.tensor(2))
            / 2
            * torch.tensor(
                [
                    [1, 1j],
                    [1j, 1],
                ],
                dtype=self.dtype,
            )
        )

    def X2P(self) -> Tensor:
        return (
            torch.sqrt(torch.tensor(2))
            / 2
            * torch.tensor(
                [
                    [1, -1j],
                    [-1j, 1],
                ],
                dtype=self.dtype,
            )
        )

    def Y2M(self) -> Tensor:
        return (
            torch.sqrt(torch.tensor(2))
            / 2
            * torch.tensor(
                [
                    [1, 1],
                    [-1, 1],
                ],
                dtype=self.dtype,
            )
        )

    def Y2P(self) -> Tensor:
        return (
            torch.sqrt(torch.tensor(2))
            / 2
            * torch.tensor(
                [
                    [1, -1],
                    [1, 1],
                ],
                dtype=self.dtype,
            )
        )

    def RX(self, theta: Tensor) -> Tensor:
        theta = self.as_tensor(theta)
        theta = torch.div(theta, 2).reshape(-1)
        return (
            torch.cat(
                (
                    torch.cos(theta),
                    -1j * torch.sin(theta),
                    -1j * torch.sin(theta),
                    torch.cos(theta),
                ),
            )
            .reshape(2, 2)
            .type(self.dtype)
        )

    def RY(self, theta: Tensor) -> Tensor:
        theta = self.as_tensor(theta)
        theta = torch.div(theta, 2).reshape(-1)
        return (
            torch.cat(
                (
                    torch.cos(theta),
                    -1 * torch.sin(theta),
                    torch.sin(theta),
                    torch.cos(theta),
                ),
            )
            .reshape(2, 2)
            .type(self.dtype)
        )

    def RZ(self, theta: Tensor) -> Tensor:
        theta = self.as_tensor(theta)
        theta = torch.div(theta, 2).reshape(-1)
        return (
            torch.cat(
                (
                    torch.exp(-1j * theta),
                    torch.zeros(1),
                    torch.zeros(1),
                    torch.exp(1j * theta),
                ),
            )
            .reshape(2, 2)
            .type(self.dtype)
        )

    def RXY(self, theta: Tensor, phi: Tensor) -> Tensor:
        theta = self.as_tensor(theta)
        phi = self.as_tensor(phi)
        theta = torch.div(theta, 2).reshape(-1)
        phi = phi.reshape(-1)

        return (
            torch.cat(
                (
                    torch.cos(theta),
                    -1j * torch.exp(-1j * phi) * torch.sin(theta),
                    -1j * torch.exp(1j * phi) * torch.sin(theta),
                    torch.cos(theta),
                ),
            )
            .reshape(2, 2)
            .type(self.dtype)
        )

    def get_zero_state(self, qnum: int) -> Tensor:
        state = torch.zeros(
            torch.pow(torch.tensor(2), torch.tensor(qnum)),
            dtype=self.dtype,
        )
        state[0] = 1
        return state

    def reshape(self, state: Tensor, shape: Sequence) -> Tensor:
        return torch.reshape(state, shape)

    def ravel(self, state: Tensor) -> Tensor:
        return torch.ravel(state)

    def stack(self, states: Sequence[Tensor], axis: int = 0) -> Tensor:
        return torch.stack(states, axis=axis)

    def real(self, state: Tensor) -> Tensor:
        return torch.real(state)

    def sum(self, state: Tensor, axis: int | None = None) -> Tensor:
        return torch.sum(state, axis=axis)

    def conj(self, state: Tensor) -> Tensor:
        return torch.conj(state)

    def sqrt(self, state: Tensor) -> Tensor:
        return torch.sqrt(state)

    def copy(self, state: Tensor) -> Tensor:
        return state.clone()

    def dot(self, state1: Tensor, state2: Tensor) -> Tensor:
        state1 = torch.as_tensor(state1)
        state2 = torch.as_tensor(state2).type_as(state1)
        return torch.dot(state1, state2)



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""A connector to use Neural Networks as PyTorch modules.
This part code is inspired by IBM qiskit.
For more information, check out:https://qiskit.org/
"""

from __future__ import annotations

from typing import Any

import numpy as np

from isqtools.neural_networks import NeuralNetwork

try:
    from torch import Tensor, einsum, sparse_coo_tensor
    from torch.autograd import Function
    from torch.nn import Module
    from torch.nn import Parameter as TorchParam
except Exception:
    Module = object
    Function = object
    Tensor = Any


class TorchWrapper(Module):
    """Connects a Quantum Neural Network to PyTorch."""

    class _TorchNNFunction(Function):
        @staticmethod
        def forward(
            ctx: Any,
            input_data: Tensor,
            weights: Tensor,
            neural_network: NeuralNetwork,
            sparse: bool,
        ) -> Tensor:
            """Forward pass computation.

            Args:
                ctx: The context to be passed to the backward pass.
                input_data: The input data.
                weights: The weights.
                neural_network: The neural network to be connected.
                sparse: Indicates whether to use sparse output or not.

            Returns:
                The resulting value of the forward pass.

            Raises:
                ValueError: Invalid input data.

            """

            # validate input shape
            if input_data.shape[-1] != neural_network.num_inputs:
                raise ValueError(
                    f"Invalid input dimension! Received {input_data.shape} and "
                    + f"expected input compatible to {neural_network.num_inputs}"
                )

            ctx.neural_network = neural_network
            ctx.sparse = sparse
            ctx.save_for_backward(input_data, weights)

            # Detach the tensors and move it to CPU as we need numpy array to compute gradients
            # of the quantum neural network. If the tensors are on CPU already this does nothing.
            # Some other tensors down below are also moved to CPU for computations.
            result = neural_network.forward(
                input_data.detach().cpu().numpy(),
                weights.detach().cpu().numpy(),
            )

            result_tensor = Tensor(result)

            # if the input was not a batch, then remove the batch-dimension from the result,
            # since the neural network will always treat input as a batch and cast to a
            # single-element batch if no batch is given and PyTorch does not follow this
            # convention.
            if len(input_data.shape) == 1:
                result_tensor = result_tensor[0]

            # place the resulting tensor back to the device where input data is stored
            result_tensor = result_tensor.to(input_data.device)

            return result_tensor

        @staticmethod
        def backward(ctx: Any, grad_output: Tensor) -> tuple:  # type: ignore
            """Backward pass computation.

            Args:
                ctx: context
                grad_output: previous gradient

            Raises:
                ValueError: Invalid input data.

            Returns:
                gradients for the first two arguments and None for the others

            """

            # get context data
            input_data, weights = ctx.saved_tensors
            neural_network = ctx.neural_network

            # if sparse output is requested return None, since PyTorch does not support it yet.
            if neural_network.sparse and ctx.sparse:
                return None, None, None, None

            # validate input shape
            if input_data.shape[-1] != neural_network.num_inputs:
                raise ValueError(
                    f"Invalid input dimension! Received {input_data.shape} and "
                    + f" expected input compatible to {neural_network.num_inputs}"
                )

            # ensure same shape for single observations and batch mode
            if len(grad_output.shape) == 1:
                grad_output = grad_output.view(1, -1)

            # evaluate QNN gradient
            input_grad, weights_grad = neural_network.backward(
                input_data.detach().cpu().numpy(),
                weights.detach().cpu().numpy(),
            )
            if input_grad is not None:
                if neural_network.sparse:
                    input_grad = sparse_coo_tensor(input_grad.coords, input_grad.data)

                    # cast to dense here, since PyTorch does not support sparse output yet.
                    # this should only happen if the network returns sparse output but the
                    # connector is configured to return dense output.
                    input_grad = (
                        input_grad.to_dense()
                    )  # this should be eventually removed
                    input_grad = input_grad.to(grad_output.dtype)
                else:
                    input_grad = Tensor(input_grad).to(grad_output.dtype)

                # Takes gradients from previous layer in backward pass (i.e. later layer in forward
                # pass) j for each observation i in the batch. Multiplies this with the gradient
                # from this point on backwards with respect to each input k. Sums over all j
                # to get total gradient of output w.r.t. each input k and batch index i.
                # This operation should preserve the batch dimension to be able to do back-prop in
                # a batched manner.
                input_grad = einsum(
                    "ij,ijk->ik", grad_output.detach().cpu(), input_grad
                )

                # place the resulting tensor to the device where they were stored
                input_grad = input_grad.to(input_data.device)

            if weights_grad is not None:
                if neural_network.sparse:
                    weights_grad = sparse_coo_tensor(
                        weights_grad.coords, weights_grad.data
                    )

                    # cast to dense here, since PyTorch does not support sparse output yet.
                    # this should only happen if the network returns sparse output but the
                    # connector is configured to return dense output.
                    weights_grad = (
                        weights_grad.to_dense()
                    )  # this should be eventually removed
                    weights_grad = weights_grad.to(grad_output.dtype)
                else:
                    weights_grad = Tensor(weights_grad).to(grad_output.dtype)

                # Takes gradients from previous layer in backward pass (i.e. later layer in forward
                # pass) j for each observation i in the batch. Multiplies this with the gradient
                # from this point on backwards with respect to each parameter k. Sums over all i and
                # j to get total gradient of output w.r.t. each parameter k.
                # The weights' dimension is independent of the batch size.
                weights_grad = einsum(
                    "ij,ijk->k", grad_output.detach().cpu(), weights_grad
                )

                # place the resulting tensor to the device where they were stored
                weights_grad = weights_grad.to(weights.device)

            # return gradients for the first two arguments and None for the others (i.e. qnn/sparse)
            return input_grad, weights_grad, None, None

    def __init__(
        self,
        neural_network: NeuralNetwork,
        initial_weights: np.ndarray | Tensor | None = None,
        sparse: bool | None = None,
    ):
        """
        Args:
            neural_network: The neural network to be connected to PyTorch. Remember
                    that ``input_gradients``  must be set to ``True`` in the neural network
                    initialization before passing it to the ``TorchConnector`` for the gradient
                    computations to work properly during training.
            initial_weights: The initial weights to start training the network. If this is None,
                the initial weights are chosen uniformly at random from [-1, 1].
            sparse: Whether this connector should return sparse output or not. If sparse is set
                to None, then the setting from the given neural network is used. Note that sparse
                output is only returned if the underlying neural network also returns sparse output,
                otherwise it will be dense independent of the setting. Also note that PyTorch
                currently does not support sparse back propagation, i.e., if sparse is set to True,
                the backward pass of this module will return None.
        """
        super().__init__()
        self._neural_network = neural_network
        self._sparse = sparse

        weight_param = TorchParam(Tensor(neural_network.num_weights))
        # Register param. in graph following PyTorch naming convention
        self.register_parameter("weight", weight_param)
        # If `weight_param` is assigned to `self._weights` after registration,
        # it will not be re-registered, and we can keep the private var. name
        # "_weights" for compatibility. The alternative, doing:
        # `self._weights = TorchParam(Tensor(neural_network.num_weights))`
        # would register the parameter with the name "_weights".
        self._weights = weight_param

        if initial_weights is None:
            self._weights.data.uniform_(-1, 1)
        else:
            self._weights.data = Tensor(initial_weights)

    @property
    def neural_network(self) -> NeuralNetwork:
        """Returns the underlying neural network."""
        return self._neural_network

    @property
    def weight(self) -> Tensor:
        """Returns the weights of the underlying network."""
        return self._weights

    @property
    def sparse(self) -> bool | None:
        """Returns whether this connector returns sparse output or not."""
        return self._sparse

    def forward(self, input_data: Tensor | None = None) -> Tensor:
        """Forward pass.

        Args:
            input_data: data to be evaluated.

        Returns:
            Result of forward pass of this model.
        """
        input_ = input_data if input_data is not None else Tensor([])
        return TorchWrapper._TorchNNFunction.apply(
            input_, self._weights, self._neural_network, self._sparse
        )

    def extra_repr(self) -> str:
        return (
            f"num_inputs={self._neural_network._num_inputs}, "
            f"num_weights={self._neural_network._num_weights}, "
            f"output_shape={self._neural_network._output_shape}"
        )



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Simple visualization of quantum circuits. These codes refer to
https://github.com/microsoft/quantum-viz.js/
"""

from collections import defaultdict
from copy import deepcopy

import matplotlib.pyplot as plt
from matplotlib import patches

single_gate = [
    "H",
    "X",
    "Y",
    "Z",
    "S",
    "T",
    "SD",
    "TD",
    "X2M",
    "X2P",
    "Y2M",
    "Y2P",
    "M",
]
multi_gate = ["CZ", "CY", "CX", "CNOT"]
theta_gate = ["RX", "RY", "RZ"]


class DrawError(Exception):
    """Draw error."""


def get_continuous_gate(opt, qlist):
    qindex = {}
    for q, i in zip(qlist, range(len(qlist))):
        qindex[q] = i
    idx = []
    idx_map = {}
    for q in opt.targets:
        idx.append(qindex[q])
        idx_map[qindex[q]] = q
    idx.append(10**7)
    idx.sort()
    continuous = []
    l, r = 0, 1
    while r < len(idx):
        if idx[r] != idx[r - 1] + 1:
            continuous.append(idx[l:r])
            l = r
        r += 1

    continuous_opt = []
    for p in continuous:
        new_opt = deepcopy(opt)
        new_opt.targets = [idx_map[i] for i in p]
        continuous_opt.append(new_opt)
    return continuous_opt


class Operation:
    def __init__(self, **kwargs) -> None:
        """create operation

        kwargs:
            gate(str): quantum gate name
            targets(list): target qubit list
            isMeasurement(bool): measurement or not
            param(str): gate's parameters
            controls(list): control qubit list
            style(dict): operation style, maybe you can change color or fontsize
        """
        self.gate = kwargs.get("gate", "")
        self.targets = kwargs.get("targets", [])
        self.isMeasurement = kwargs.get("isMeasurement", False)
        self.param = kwargs.get("param", "")
        self.controls = kwargs.get("controls", [])
        self.style = kwargs.get("style", {})

        if not self.gate:
            raise DrawError("gate can not be empty")
        if not self.targets:
            raise DrawError("target qubit can not be empty")

        qubits = self.targets + self.controls
        if len(set(qubits)) != len(qubits):
            raise DrawError("qubit reuse")
        if self.isMeasurement:
            if len(self.targets) > 1:
                raise DrawError("can only measure one qubit")
            if len(self.controls) > 0:
                raise DrawError("can not use control measure")

    def __repr__(self) -> str:
        return f"Operation(gate={self.gate}, targets={self.targets}, controls={self.controls}, isMeasurement={self.isMeasurement}, param={self.param}, style={self.style})"

    def to_dict(self):
        d = {
            "gate": self.gate,
            "targets": self.targets,
        }
        if self.controls:
            d["controls"] = self.controls
        if self.isMeasurement:
            d["isMeasurement"] = True
        if self.param:
            d["param"] = self.param
        if self.style:
            d["style"] = self.style
        return d


class Circuit:
    def __init__(self, **kwargs) -> None:
        """create circuit

        kwargs:
            qubits(list): qubit list
            operations(list): operation list, element can be dict or Operation
            max_gate_num(int): circuit depth, which will cut circuit, default is 20
            style(dict): circuit style, maybe you can change color or fontsize
            show_param(bool): wheather show parameter, like Rx(0.4)
            optimized(bool): wheather reorder circuit, avoid gate crossings
            qubits_labels(dict): qubit label, default is qubit name
            show_label(bool): wheather show lables
            gate_group(list): Use a dashed box to draw out a group of gates
        """
        self.qlist = kwargs.get("qubits", [])
        if not self.qlist:
            raise "qubit list can not be empty"
        self.qmap = {}
        self.qmap_inv = {}
        for q, i in zip(self.qlist, range(len(self.qlist))):
            self.qmap[q] = i
            self.qmap_inv[i] = q

        self.max_circuit_depth = kwargs.get("max_circuit_depth", 20)
        self.style = kwargs.get("style", {})
        self.show_param = kwargs.get("show_param", False)
        self.optimized = kwargs.get("optimized", False)
        self.qubits_labels = kwargs.get("qubits_labels", {})
        self.show_label = kwargs.get("show_label", True)
        self.gate_group = kwargs.get("gate_group", [])

        for q in self.qlist:
            if q not in self.qubits_labels:
                self.qubits_labels[q] = q

        if "operations" not in kwargs:
            raise DrawError("operations can not be empty")

        self.operations = []
        for opt in kwargs["operations"]:
            if isinstance(opt, Operation):
                self.operations.append(opt)
            elif isinstance(opt, dict):
                self.operations.append(Operation(**opt))
            else:
                raise DrawError("operation type error")
        self.build_seqs()

        self.split_gate_group = []
        m = len(self.seqs) - 1
        max_t = self.max_circuit_depth * (m) + len(self.seqs[m])
        for a, b, c, d in self.gate_group:
            if a > c or b > d:
                raise DrawError(
                    "gate group position error, need top-left and bottom-right"
                )
            if a < 0 or c >= len(self.qlist):
                raise DrawError(
                    "gate group position error, exceeding the number of qubits"
                )
            d = min(d, max_t - 1)
            while b <= d:
                nb = b + self.max_circuit_depth - b % self.max_circuit_depth
                self.split_gate_group.append([a, b, c, min(d, nb - 1)])
                b = nb

    def get_path_qubits(self, qubits):
        idx = [self.qmap[q] for q in qubits]
        s, t = min(idx), max(idx)
        path_qubits = [self.qmap_inv[i] for i in range(s, t + 1)]
        return path_qubits

    def build_seqs(self):
        node_time = defaultdict(int)
        occupy_time = defaultdict(set)

        def get_time(i):
            while node_time[i] in occupy_time[i]:
                node_time[i] += 1
            return node_time[i]

        seq_time = 0
        seqs = defaultdict(dict)
        max_t = self.max_circuit_depth
        for opt in self.operations:
            qubits = opt.targets + opt.controls
            path_qubits = qubits.copy()
            if self.optimized:
                path_qubits = self.get_path_qubits(qubits)
            t = max([get_time(q) for q in path_qubits])
            if (t % max_t) not in seqs[t // max_t]:
                seqs[t // max_t][t % max_t] = []
            seqs[t // max_t][t % max_t].append(opt)
            if self.optimized:
                for q in path_qubits:
                    occupy_time[q].add(t)
            for q in qubits:
                node_time[q] = t + 1
            seq_time = max(seq_time, t)
        self.seqs = seqs

    def to_dict(self):
        d = {
            "qubits": self.qlist,
            "operations": [opt.to_dict() for opt in self.operations],
        }
        if self.max_circuit_depth:
            d["max_circuit_depth"] = self.max_circuit_depth
        if self.style:
            d["style"] = self.style
        if self.show_param:
            d["show_param"] = self.show_param
        if self.gate_group:
            d["gate_group"] = self.gate_group
        d["show_label"] = self.show_label
        d["qubits_labels"] = self.qubits_labels
        return d

    def parse_to_latex(self):
        qcircuit = defaultdict(list)

        for m in self.seqs:
            seqs = self.seqs[m]
            for idx in range(len(seqs)):
                for q in self.qlist:
                    qcircuit[q].append("& \qw")
                for opt in seqs[idx]:
                    if opt.isMeasurement:
                        qcircuit[opt.targets[0]][-1] = "& \meter"
                    else:
                        continuous_opt = get_continuous_gate(opt, self.qlist)
                        for copt in continuous_opt:
                            if copt.gate == "X" and len(copt.controls) > 0:
                                for i, q in enumerate(copt.targets):
                                    qcircuit[q][-1] = "& \\targ"
                                    if i > 0:
                                        qcircuit[q][
                                            -1
                                        ] += f"\qwx[{-1 * (self.qmap[q] - self.qmap[copt.targets[i-1]])}]"
                            else:
                                gate_name = copt.gate
                                if self.show_param and copt.param:
                                    gate_name += copt.param
                                if len(copt.targets) > 1:
                                    qcircuit[copt.targets[0]][
                                        -1
                                    ] = f"& \multigate{{{len(copt.targets) - 1}}}{{{gate_name}}}"
                                    for i in range(1, len(copt.targets)):
                                        qcircuit[copt.targets[i]][
                                            -1
                                        ] = f"& \ghost{{{gate_name}}}"
                                else:
                                    qcircuit[copt.targets[0]][
                                        -1
                                    ] = f"& \gate{{{gate_name}}}"

                        if opt.controls:
                            for q in opt.controls:
                                qcircuit[q][-1] = "& \ctrl"
                            interval = [
                                (
                                    self.qmap[o.targets[0]],
                                    self.qmap[o.targets[-1]],
                                    "T",
                                )
                                for o in continuous_opt
                            ]
                            interval += [
                                (self.qmap[q], self.qmap[q], "C") for q in opt.controls
                            ]
                            interval.sort()
                            for i, (s, t, ch) in enumerate(interval):
                                if ch == "C":
                                    if i > 0:
                                        qcircuit[self.qmap_inv[s]][
                                            -1
                                        ] += f"{{{-1*(s - interval[i-1][1])}}}"
                                    else:
                                        qcircuit[self.qmap_inv[s]][
                                            -1
                                        ] += f"{{{(interval[i+1][0] - t)}}}"
                                else:
                                    if i > 0:
                                        qcircuit[self.qmap_inv[s]][
                                            -1
                                        ] += f"\qwx[{-1*(s - interval[i-1][1])}]"
                                    else:
                                        qcircuit[self.qmap_inv[s]][
                                            -1
                                        ] += f"\qwx[{(interval[i+1][0] - t)}]"
                        else:
                            for i in range(1, len(continuous_opt)):
                                q1, q2 = (
                                    continuous_opt[i].targets[0],
                                    continuous_opt[i - 1].targets[-1],
                                )
                                s, t = self.qmap[q1], self.qmap[q2]
                                qcircuit[q1][-1] += f"\qwx[{-1*(s - t)}]"

        return qcircuit

    def get_latex(self):
        """get latex from circuit"""
        qcircuit = self.parse_to_latex()
        latex_text = r"""\documentclass[border=2px]{standalone}
\usepackage[braket, qm]{qcircuit}
\begin{document}
\Qcircuit @C=1em @R=1em {
"""
        i = 0
        n = len(qcircuit[self.qlist[0]])
        m = n + 3
        if not self.show_label:
            m = n + 1
        while i < n:
            j = min(i + self.max_circuit_depth, n)
            for q in self.qlist:
                label = f"\\nghost{{{self.qubits_labels[q]}}} & \lstick{{{self.qubits_labels[q]}}}"
                if not self.show_label:
                    label = ""
                latex_text += f'{label} {"".join(qcircuit[q][i:j])} & \qw \\\\\n'
            latex_text += f'{"".join(["&"] * m)} \\\\\n'
            i += self.max_circuit_depth

        if self.split_gate_group:
            x_offset = 2
            if self.show_label:
                x_offset += 2
            latex_text += "\\ "
            for a, b, c, d in self.split_gate_group:
                y_offset = (b // self.max_circuit_depth) * (len(self.qlist) + 1) + 1
                b, d = b % self.max_circuit_depth, d % self.max_circuit_depth
                latex_text += f"\gategroup{{{a+y_offset}}}{{{b+x_offset}}}{{{c+y_offset}}}{{{d+x_offset}}}{{.7em}}{{--}} "
        latex_text += """}
\end{document}
"""
        return latex_text


def normalize(isq_ir):
    ir_item = isq_ir.split(" ")
    qid1 = ir_item[1]
    if ir_item[0] in single_gate:
        if ir_item[0] == "M":
            return Operation(gate=ir_item[0], targets=[qid1], isMeasurement=True)
        return Operation(gate=ir_item[0], targets=[qid1])
    elif ir_item[0] in multi_gate:
        qid2 = ir_item[2]
        if ir_item[0] in ["CX", "CNOT"]:
            return Operation(gate="X", targets=[qid2], controls=[qid1])
        if ir_item[0] == "CZ":
            return Operation(gate="Z", targets=[qid2], controls=[qid1])
        if ir_item[0] == "CY":
            return Operation(gate="Y", targets=[qid2], controls=[qid1])
    elif ir_item[0] in theta_gate:
        return Operation(
            gate=ir_item[0], targets=[qid1], param=f'({",".join(ir_item[2:])})'
        )


def get_circuit(isq_ir, **kwargs):
    """get circuit from isq ir

    Args:
        isq_ir (str): isq ir

    Returns: Circuit
    """
    isq_ir = isq_ir.strip()
    qubits = set()
    operations = []
    for ir in isq_ir.split("\n"):
        ir = ir.strip()
        if ir == "":
            continue
        opt = normalize(ir)
        operations.append(opt)
        for q in opt.targets:
            qubits.add(q)
        for q in opt.controls:
            qubits.add(q)
    qlist = list(qubits)
    qlist.sort()
    return Circuit(qubits=qlist, operations=operations, **kwargs)


class Drawer:
    def __init__(self) -> None:
        self.box_length = 1.0
        self.box_width = 1.0
        self.pad = 0.3
        self.ctrl_rad = 0.1
        self.circ_rad = 0.3
        self.xinterval = 0.3
        self.yinterval = 0.3

    def init_style(self):
        self.line_options = {"color": "black", "linewidth": 1, "zorder": 2}
        self.ctrl_line_options = {"color": "black", "linewidth": 1, "zorder": 1}
        self.group_line_options = {
            "color": "black",
            "linewidth": 1,
            "zorder": 2,
            "linestyle": (0, (10, 10)),
        }
        self.backgroud_line_options = {
            "color": "black",
            "linewidth": 1,
            "zorder": 1,
        }
        self.dash_line_options = {
            "color": "black",
            "linewidth": 1,
            "zorder": 1,
            "linestyle": "--",
        }
        self.text_options = {
            "zorder": 3,
            "ha": "center",
            "va": "center",
            "fontsize": int(self.box_width * 10),
        }
        self.box_options = {
            "facecolor": "white",
            "edgecolor": "black",
            "linewidth": 1,
            "zorder": 2,
        }
        self.arc_options = {"linewidth": 1, "zorder": 3}
        self.arrow_options = {
            "facecolor": "black",
            "edgecolor": "black",
            "linewidth": 1,
            "zorder": 3,
        }
        self.boxstyle = "square, pad=0.2"

    def set_circuit_style(self):
        for attr in self.__dict__:
            if attr.endswith("options"):
                self.__dict__[attr].update(self.circuit.style.get(attr, {}))

    def set_opt_style(self, opt):
        self.ori_box_options = deepcopy(self.box_options)
        self.ori_text_options = deepcopy(self.text_options)
        if opt.style:
            self.box_options.update(opt.style.get("box_options", {}))
            self.text_options.update(opt.style.get("text_options", {}))

    def recover_style(self):
        self.box_options = deepcopy(self.ori_box_options)
        self.text_options = deepcopy(self.ori_text_options)

    def plot(self, circuit, **kwargs):
        """plot circuit

        Args:
            circuit (str | dict | Circuit): input circuit which can be isq ir or Circuit
        """

        # change ir to circuit
        if not isinstance(circuit, Circuit):
            if isinstance(circuit, str):
                circuit = get_circuit(circuit, **kwargs)
            elif isinstance(circuit, dict):
                circuit = Circuit(**circuit)
            else:
                raise DrawError("circuit type error")

        self.show_param = circuit.show_param
        self.circuit = circuit
        self.init_style()
        self.set_circuit_style()

        self.qmap = {}
        for i, q in enumerate(self.circuit.qlist):
            self.qmap[q] = i * (self.box_length + self.yinterval)

        self.get_figure_size()
        self.set_figure()
        self.draw_circuit()
        self.draw_groups()

    def layer_height(self):
        return (
            len(self.circuit.qlist) * (self.box_length + self.yinterval)
            + self.yinterval
        )

    def get_figure_size(self):
        ax_seqs = self.circuit.seqs
        width = 0
        height = 1
        for m in ax_seqs:
            n = len(ax_seqs[m])
            ax_width = self.box_width
            for i in range(n):
                ax_width += self.get_width_of_seqs(ax_seqs[m][i])
                ax_width += self.xinterval
            ax_width -= self.xinterval
            width = max(width, ax_width)
            height += self.layer_height()
        width += 1.5
        self.figure_width = width
        self.figure_height = height

    def set_figure(self):
        self.fg = plt.figure(figsize=(self.figure_width / 2, self.figure_height / 2))
        self.ax = self.fg.add_axes(
            [0, 0, 1, 1],
            xlim=(0, self.figure_width),
            ylim=(-self.box_length, self.figure_height - 1),
            xticks=[],
            yticks=[],
        )
        self.ax.axis("off")
        self.ax.invert_yaxis()

    def draw_circuit(self):
        ax_y_offset = 0
        qnum = len(self.circuit.qlist)

        self.center = defaultdict(float)

        for m in range(len(self.circuit.seqs)):
            seqs = self.circuit.seqs[m]
            center = 1.5
            for i in range(len(seqs)):
                # get sequence center
                in_width = self.get_width_of_seqs(seqs[i])
                center += (in_width - self.box_width) / 2
                for opt in seqs[i]:
                    self.set_opt_style(opt)
                    self.add_opt(opt, center)
                    self.recover_style()
                self.center[m * self.circuit.max_circuit_depth + i] = center
                center += (in_width - self.box_width) / 2
                center += self.box_width + self.xinterval

            # plot qubit lines
            start_x = 0.5
            label_text_options = deepcopy(self.text_options)
            label_text_options["ha"] = "right"
            for i in range(qnum):
                y = i * (self.box_length + self.yinterval) + ax_y_offset
                if self.circuit.show_label:
                    self.ax.text(
                        start_x - 0.1,
                        self.qmap[self.circuit.qlist[i]],
                        f"{self.circuit.qubits_labels[self.circuit.qlist[i]]}",
                        **label_text_options,
                    )
                line = plt.Line2D(
                    (start_x, self.figure_width - 1),
                    (y, y),
                    **self.backgroud_line_options,
                )
                self.ax.add_line(line)

            ax_y_offset += self.layer_height()
            for qubit in self.qmap:
                self.qmap[qubit] += self.layer_height()

    def draw_groups(self):
        for a, b, c, d in self.circuit.split_gate_group:
            m = b // self.circuit.max_circuit_depth
            x1, x2 = self.center[b], self.center[d]
            b, d = (
                b % self.circuit.max_circuit_depth,
                d % self.circuit.max_circuit_depth,
            )
            w1 = self.get_width_of_seqs(self.circuit.seqs[m][b])
            w2 = self.get_width_of_seqs(self.circuit.seqs[m][d])
            x1 -= w1 / 2
            x2 += w2 / 2
            y1 = (
                m * self.layer_height()
                + a * (self.box_length + self.yinterval)
                - self.box_length / 2
            )
            y2 = (
                m * self.layer_height()
                + c * (self.box_length + self.yinterval)
                + self.box_length / 2
            )
            for x in [x1, x2]:
                line = plt.Line2D((x, x), (y1, y2), **self.group_line_options)
                self.ax.add_line(line)
            for y in [y1, y2]:
                line = plt.Line2D((x1, x2), (y, y), **self.group_line_options)
                self.ax.add_line(line)

    def get_width_of_seqs(self, seqs):
        width = self.box_width
        for opt in seqs:
            width = max(width, self.get_width_of_opt(opt))
        return width

    def get_width_of_opt(self, opt):
        width = self.box_width
        m = len(opt.gate)
        if self.show_param:
            m = max(m, len(opt.param))
        width = max(width, 1 * (m / 6))
        # print(opt, width)
        return width

    def box(self, x, y, box_width, box_length):
        """build a box

        Args:
            x (float): center point's x
            y (float): center point's y
            box_width (float): box width
            box_length (float): box length
        """
        pad = self.pad

        x_loc = x - box_width / 2.0 + pad
        y_loc = y - box_length / 2.0 + pad

        box = patches.FancyBboxPatch(
            (x_loc, y_loc),
            box_width - 2 * pad,
            box_length - 2 * pad,
            boxstyle=self.boxstyle,
            **self.box_options,
        )
        self.ax.add_patch(box)

    def add_ctrl_line(self, y1, y2, seq_t):
        line = plt.Line2D((seq_t, seq_t), (y1, y2), **self.ctrl_line_options)
        self.ax.add_line(line)

    def add_continuous_gate(self, opt, seq_t):
        """draw gate

        Args:
            opt (Operation): operation which need be draw
            seq_t (float): center point's x
        """

        if len(opt.controls) > 0 and opt.gate == "X":
            for i, qubit in enumerate(opt.targets):
                self.add_ctrl_x_gate(qubit, seq_t)
                if i > 0:
                    self.add_ctrl_line(
                        self.qmap[opt.targets[i - 1]], self.qmap[qubit], seq_t
                    )
        else:
            box_length = (
                self.qmap[opt.targets[-1]] - self.qmap[opt.targets[0]] + self.box_length
            )
            y = (self.qmap[opt.targets[-1]] + self.qmap[opt.targets[0]]) / 2

            box_width = self.get_width_of_opt(opt)

            self.box(seq_t, y, box_width, box_length)

            if self.show_param and opt.param != "":
                self.ax.text(
                    seq_t,
                    y + self.box_length / 8,
                    opt.param,
                    **self.text_options,
                )

                self.ax.text(
                    seq_t,
                    y - self.box_length / 8,
                    opt.gate,
                    **self.text_options,
                )
            else:
                self.ax.text(seq_t, y, opt.gate, **self.text_options)

    def add_ctrl_x_gate(self, qubit, seq_t):
        y = self.qmap[qubit]
        circ_rad = self.circ_rad
        new_options = {
            "facecolor": "white",
            "edgecolor": "black",
            "linewidth": 2,
            "zorder": 2,
        }
        target_circ = plt.Circle((seq_t, y), radius=circ_rad, **new_options)
        self.line_options["zorder"] = 3
        self.line_options["linewidth"] = 2
        target_v = plt.Line2D(
            (seq_t, seq_t), (y - circ_rad, y + circ_rad), **self.line_options
        )
        target_h = plt.Line2D(
            (seq_t - circ_rad, seq_t + circ_rad), (y, y), **self.line_options
        )
        self.line_options["zorder"] = 2
        self.line_options["linewidth"] = 1
        self.ax.add_patch(target_circ)
        self.ax.add_line(target_v)
        self.ax.add_line(target_h)

    def add_opt(self, opt, seq_t):
        if opt.isMeasurement:
            self.add_measure(opt, seq_t)
            return

        continuous_opt = get_continuous_gate(opt, self.circuit.qlist)
        for gate in continuous_opt:
            self.add_continuous_gate(gate, seq_t)
            if not opt.controls:
                for i in range(1, len(continuous_opt)):
                    q1, q2 = (
                        continuous_opt[i - 1].targets[-1],
                        continuous_opt[i].targets[0],
                    )
                    y1, y2 = self.qmap[q1], self.qmap[q2]
                    line = plt.Line2D(
                        (seq_t, seq_t), (y1, y2), **self.dash_line_options
                    )
                    self.ax.add_line(line)

        if opt.controls:
            ctrl_rad = self.ctrl_rad
            for qubit in opt.controls:
                circ_ctrl = plt.Circle(
                    (seq_t, self.qmap[qubit]),
                    radius=ctrl_rad,
                    **self.line_options,
                )
                self.ax.add_patch(circ_ctrl)

            interval = [
                (self.qmap[o.targets[0]], self.qmap[o.targets[-1]], "T")
                for o in continuous_opt
            ]
            interval += [(self.qmap[q], self.qmap[q], "C") for q in opt.controls]
            interval.sort()
            for i in range(1, len(interval)):
                y1, y2 = interval[i - 1][1], interval[i][0]
                self.add_ctrl_line(y1, y2, seq_t)

    def add_measure(self, opt, seq_t):
        """draw measure

        Args:
            opt (Operation): operation which need be draw
            seq_t (float): center point's x
        """
        y = self.qmap[opt.targets[0]]

        self.box(seq_t, y, self.box_width, self.box_length)

        box_length = self.box_length

        arc = patches.Arc(
            (seq_t, y + box_length / 5),
            0.6 * box_length,
            0.55 * box_length,
            theta1=180,
            theta2=0,
            **self.arc_options,
        )
        self.ax.add_patch(arc)

        arrow_start_x = seq_t - 0.165 * box_length
        arrow_start_y = y + 0.25 * box_length
        arrow_width = 0.35 * box_length
        arrow_height = -0.35 * box_length

        self.ax.arrow(
            arrow_start_x,
            arrow_start_y,
            arrow_width,
            arrow_height,
            head_width=box_length / 8.0,
            **self.arrow_options,
        )

    def save(self, filename):
        self.fg.savefig(f"{filename}", bbox_inches="tight")
